from contextlib import contextmanager
from typing import Text, Dict, Optional, List

import ibm_boto3
from ibm_botocore.client import ClientError
import pandas as pd
import duckdb

from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb_sql_parser.ast import Select, Identifier, Insert, Star, Constant

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)

from mindsdb.integrations.libs.api_handler import APIResource, APIHandler
from mindsdb.integrations.utilities.sql_utils import FilterCondition, FilterOperator

logger = log.getLogger(__name__)


class ListFilesTable(APIResource):
    def list(
        self,
        targets: List[str] = None,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        *args,
        **kwargs,
    ) -> pd.DataFrame:

        buckets = None
        for condition in conditions:
            if condition.column == "bucket":
                if condition.op == FilterOperator.IN:
                    buckets = condition.value
                elif condition.op == FilterOperator.EQUAL:
                    buckets = [condition.value]
                condition.applied = True

        data = []
        for obj in self.handler.get_objects(limit=limit, buckets=buckets):
            path = obj["Key"]
            if obj["Filename"].split(".")[1] in self.handler.supported_file_formats:
                item = {
                    "path": path,
                    "bucket": obj["Bucket"],
                    "name": path[path.rfind("/") + 1:],
                    "extension": path[path.rfind(".") + 1:],
                }

                data.append(item)

        return pd.DataFrame(data=data, columns=self.get_columns())

    def get_columns(self) -> List[str]:
        return ["path", "name", "extension", "bucket", "content"]


class FileTable(APIResource):
    def list(
        self, targets: List[str] = None, table_name=None, *args, **kwargs
    ) -> pd.DataFrame:
        return self.handler.read_as_table(table_name)

    def add(self, data, table_name=None):
        df = pd.DataFrame(data)
        return self.handler.add_data_to_table(table_name, df)


class IBMCloudObjectStorageHandler(APIHandler):

    name = "ibm_cos"
    supported_file_formats = ["csv", "tsv", "json", "parquet"]

    def __init__(self, name: Text, connection_data: Optional[Dict] = None, **kwargs):
        super().__init__(name)
        self.connection_data = connection_data or {}
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False
        self.thread_safe = True
        self._regions = {}

        self.bucket = self.connection_data.get("bucket")
        self._files_table = ListFilesTable(self)

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        if self.is_connected is True:
            return self.connection

        required_params = [
            "cos_hmac_access_key_id",
            "cos_hmac_secret_access_key",
            "cos_endpoint_url",
        ]
        if not all(key in self.connection_data for key in required_params):
            raise ValueError(
                "Required parameters (cos_hmac_access_key_id, cos_hmac_secret_access_key, cos_endpoint_url) must be provided."
            )

        self.connection = self._connect_ibm_boto3()
        self.is_connected = True

        return self.connection

    def _connect_ibm_boto3(self) -> ibm_boto3.client:
        config = {
            "aws_access_key_id": self.connection_data["cos_hmac_access_key_id"],
            "aws_secret_access_key": self.connection_data["cos_hmac_secret_access_key"],
            "endpoint_url": self.connection_data["cos_endpoint_url"],
        }

        client = ibm_boto3.client("s3", **config)

        if self.bucket is not None:
            client.head_bucket(Bucket=self.bucket)
        else:
            client.list_buckets()

        return client

    def disconnect(self):
        if not self.is_connected:
            return
        self.connection = None
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self._connect_ibm_boto3()
            response.success = True
        except (ClientError, ValueError) as e:
            logger.error(
                f"Error connecting to IBM COS with the given credentials, {e}!"
            )
            response.error_message = str(e)

        if response.success and need_to_close:
            self.disconnect()

        elif not response.success and self.is_connected:
            self.is_connected = False

        return response

    @contextmanager
    def _connect_duckdb(self):
        duckdb_conn = duckdb.connect(":memory:")
        duckdb_conn.execute("INSTALL httpfs")
        duckdb_conn.execute("LOAD httpfs")

        duckdb_conn.execute(
            f"SET s3_access_key_id='{self.connection_data['cos_hmac_access_key_id']}'"
        )
        duckdb_conn.execute(
            f"SET s3_secret_access_key='{self.connection_data['cos_hmac_secret_access_key']}'"
        )

        endpoint_url = self.connection_data["cos_endpoint_url"]
        if endpoint_url.startswith("https://"):
            endpoint_url = endpoint_url[len("https://"):]
        elif endpoint_url.startswith("http://"):
            endpoint_url = endpoint_url[len("http://"):]

        duckdb_conn.execute(f"SET s3_endpoint='{endpoint_url}'")
        duckdb_conn.execute("SET s3_url_style='path'")
        duckdb_conn.execute("SET s3_use_ssl=true")

        try:
            yield duckdb_conn
        finally:
            duckdb_conn.close()

    def _get_bucket(self, key):
        if self.bucket is not None:
            return self.bucket, key

        ar = key.split("/")
        return ar[0], "/".join(ar[1:])

    def read_as_table(self, key) -> pd.DataFrame:
        bucket, key = self._get_bucket(key)

        with self._connect_duckdb() as connection:

            cursor = connection.execute(f"SELECT * FROM 's3://{bucket}/{key}'")

            return cursor.fetchdf()

    def _read_as_content(self, key) -> None:
        bucket, key = self._get_bucket(key)

        client = self.connect()

        obj = client.get_object(Bucket=bucket, Key=key)
        content = obj["Body"].read()
        return content

    def add_data_to_table(self, key, df) -> None:

        bucket, key = self._get_bucket(key)

        try:
            client = self.connect()
            client.head_object(Bucket=bucket, Key=key)
        except ClientError as e:
            logger.error(f"Error querying the file {key} in the bucket {bucket}, {e}!")
            raise e

        with self._connect_duckdb() as connection:
            connection.execute(
                f"CREATE TABLE tmp_table AS SELECT * FROM 's3://{bucket}/{key}'"
            )

            connection.execute("INSERT INTO tmp_table BY NAME SELECT * FROM df")

            connection.execute(f"COPY tmp_table TO 's3://{bucket}/{key}'")

    def query(self, query: ASTNode) -> Response:
        self.connect()
        if isinstance(query, Select):
            table_name = query.from_table.parts[-1]

            if table_name == "files":
                table = self._files_table
                df = table.select(query)

                has_content = False
                for target in query.targets:
                    if (
                        isinstance(target, Identifier)
                        and target.parts[-1].lower() == "content"
                    ):
                        has_content = True
                        break
                if has_content:
                    df["content"] = df["path"].apply(self._read_as_content)
            else:
                extension = table_name.split(".")[-1]
                if extension not in self.supported_file_formats:
                    logger.error(f"The file format {extension} is not supported!")
                    raise ValueError(f"The file format {extension} is not supported!")

                table = FileTable(self, table_name=table_name)
                df = table.select(query)

            response = Response(RESPONSE_TYPE.TABLE, data_frame=df)
        elif isinstance(query, Insert):
            table_name = query.table.parts[-1]
            table = FileTable(self, table_name=table_name)
            table.insert(query)
            response = Response(RESPONSE_TYPE.OK)
        else:
            raise NotImplementedError

        return response

    def get_objects(self, limit=None, buckets=None) -> List[dict]:
        client = self.connect()
        if self.bucket is not None:
            add_bucket_to_name = False
            scan_buckets = [self.bucket]
        else:
            add_bucket_to_name = True
            resp = client.list_buckets()
            scan_buckets = [b["Name"] for b in resp["Buckets"]]

        objects = []
        for bucket in scan_buckets:
            if buckets is not None and bucket not in buckets:
                continue

            resp = client.list_objects_v2(Bucket=bucket)
            if "Contents" not in resp:
                continue

            for obj in resp["Contents"]:
                obj["Bucket"] = bucket
                obj["Filename"] = obj["Key"]
                if add_bucket_to_name:
                    obj["Key"] = f'{bucket}/{obj["Key"]}'
                objects.append(obj)
            if limit is not None and len(objects) >= limit:
                break

        return objects

    def get_tables(self) -> Response:
        supported_names = [
            f"{obj['Key']}"
            for obj in self.get_objects()
            if obj["Key"].split(".")[-1] in self.supported_file_formats
        ]

        supported_names.insert(0, "files")

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(supported_names, columns=["table_name"]),
        )

        return response

    def get_columns(self, table_name: str) -> Response:
        query = Select(
            targets=[Star()],
            from_table=Identifier(parts=[table_name]),
            limit=Constant(1),
        )

        result = self.query(query)

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                {
                    "column_name": result.data_frame.columns,
                    "data_type": [
                        str(dtype) if str(dtype) != "object" else "string"
                        for dtype in result.data_frame.dtypes
                    ],
                }
            ),
        )

        return response



from typing import Optional
from collections import OrderedDict

from mindsdb.interfaces.database.projects import ProjectController
import mindsdb.utilities.profiler as profiler
from mindsdb.utilities.config import config
from mindsdb.utilities.exception import EntityNotExistsError
from mindsdb.interfaces.database.log import LogDBController


class DatabaseController:
    def __init__(self):
        from mindsdb.interfaces.database.integrations import integration_controller

        self.integration_controller = integration_controller
        self.project_controller = ProjectController()

        self.logs_db_controller = LogDBController()
        self.information_schema_controller = None

    def delete(self, name: str, strict_case: bool = False) -> None:
        """Delete a database (project or integration) by name.

        Args:
            name (str): The name of the database to delete.
            strict_case (bool, optional): If True, the database name is case-sensitive. Defaults to False.

        Raises:
            EntityNotExistsError: If the database does not exist.
            Exception: If the database cannot be deleted.

        Returns:
            None
        """
        databases = self.get_dict()
        if name.lower() not in databases:
            raise EntityNotExistsError("Database does not exists", name)
        db_type = databases[name.lower()]["type"]
        if db_type == "project":
            project = self.get_project(name, strict_case)
            project.delete()
            return
        elif db_type == "data":
            self.integration_controller.delete(name, strict_case)
            return
        else:
            raise Exception(f"Database with type '{db_type}' cannot be deleted")

    @profiler.profile()
    def get_list(self, filter_type: Optional[str] = None, with_secrets: Optional[bool] = True):
        projects = self.project_controller.get_list()
        integrations = self.integration_controller.get_all(show_secrets=with_secrets)
        result = [
            {
                "name": "information_schema",
                "type": "system",
                "id": None,
                "engine": None,
                "visible": True,
                "deletable": False,
            },
            {"name": "log", "type": "system", "id": None, "engine": None, "visible": True, "deletable": False},
        ]
        for x in projects:
            result.append(
                {
                    "name": x.name,
                    "type": "project",
                    "id": x.id,
                    "engine": None,
                    "visible": True,
                    "deletable": x.name.lower() != config.get("default_project"),
                }
            )
        for key, value in integrations.items():
            db_type = value.get("type", "data")
            if db_type != "ml":
                result.append(
                    {
                        "name": key,
                        "type": value.get("type", "data"),
                        "id": value.get("id"),
                        "engine": value.get("engine"),
                        "class_type": value.get("class_type"),
                        "connection_data": value.get("connection_data"),
                        "visible": True,
                        "deletable": value.get("permanent", False) is False,
                    }
                )

        if filter_type is not None:
            result = [x for x in result if x["type"] == filter_type]

        return result

    def get_dict(self, filter_type: Optional[str] = None, lowercase: bool = True):
        return OrderedDict(
            (x["name"].lower() if lowercase else x["name"], {"type": x["type"], "engine": x["engine"], "id": x["id"]})
            for x in self.get_list(filter_type=filter_type)
        )

    def get_integration(self, integration_id):
        # get integration by id

        # TODO get directly from db?
        for rec in self.get_list():
            if rec["id"] == integration_id and rec["type"] == "data":
                return {"name": rec["name"], "type": rec["type"], "engine": rec["engine"], "id": rec["id"]}

    def exists(self, db_name: str) -> bool:
        return db_name.lower() in self.get_dict()

    def get_project(self, name: str, strict_case: bool = False):
        """Get a project by name.

        Args:
            name (str): The name of the project to retrieve.
            strict_case (bool, optional): If True, the project name is case-sensitive. Defaults to False.

        Returns:
            Project: The project instance matching the given name.
        """
        return self.project_controller.get(name=name, strict_case=strict_case)

    def get_system_db(self, name: str):
        if name == "log":
            return self.logs_db_controller
        elif name == "information_schema":
            from mindsdb.api.executor.controllers.session_controller import SessionController

            session = SessionController()
            return session.datahub
        else:
            raise Exception(f"Database '{name}' does not exists")

    def update(self, name: str, data: dict, strict_case: bool = False):
        """
        Updates the database with the given name using the provided data.

        Parameters:
            name (str): The name of the database to update.
            data (dict): The data to update the database with.
            strict_case (bool): if True, then name is case-sesitive

        Raises:
            EntityNotExistsError: If the database does not exist.
        """
        databases = self.get_dict(lowercase=(not strict_case))
        if not strict_case:
            name = name.lower()
        if name not in databases:
            raise EntityNotExistsError("Database does not exist.", name)

        db_type = databases[name]["type"]
        if db_type == "project":
            # Only the name of the project can be updated.
            if {"name"} != set(data):
                raise ValueError("Only the 'name' field can be updated for projects.")
            if not data["name"].islower():
                raise ValueError("New name must be in lower case.")
            self.project_controller.update(name=name, new_name=str(data["name"]))
            return

        elif db_type == "data":
            # Only the parameters (connection data) of the integration can be updated.
            if {"parameters"} != set(data):
                raise ValueError("Only the 'parameters' field can be updated for integrations.")
            self.integration_controller.modify(name, data["parameters"])
            return

        else:
            raise ValueError(f"Database with type '{db_type}' cannot be updated")



import boto3
from typing import Text, Optional


def create_amazon_bedrock_client(
    client: Text,
    aws_access_key_id: Text,
    aws_secret_access_key: Text,
    region_name: Text,
    aws_session_token: Optional[Text] = None,
) -> boto3.client:
    """
    Create an Amazon Bedrock client via boto3.

    Parameters
    ----------
    client : Text
        The type of client to create. It can be 'bedrock' or 'bedrock-runtime'.

    aws_access_key_id : Text
        The AWS access key ID.

    aws_secret_access_key : Text
        The AWS secret access key.

    region_name : Text
        The AWS region name.

    aws_session_token : Text, Optional
        The AWS session token. Optional, but required for temporary security credentials.

    Returns
    -------
    boto3.client
        Amazon Bedrock client.
    """
    if client not in ["bedrock", "bedrock-runtime"]:
        raise ValueError("The client must be 'bedrock' or 'bedrock-runtime'")

    return boto3.client(
        client,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
        aws_session_token=aws_session_token,
    )



import dill
import pandas as pd
from mindsdb.integrations.libs.base import BaseMLEngine
from typing import Dict, Optional
from type_infer.api import infer_types
from flaml import AutoML


class FLAMLHandler(BaseMLEngine):
    name = "FLAML"

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        if args is None:
            args = {}

        if df is not None:
            target_dtype = infer_types(df, 0).to_dict()["dtypes"][target]
            model = AutoML(verbose=0)

            if target_dtype in ['binary', 'categorical', 'tags']:
                model.fit(X_train=df.drop(columns=[target]),
                          y_train=df[target],
                          task='classification',
                          **args.get('using'))

            elif target_dtype in ['integer', 'float', 'quantity']:
                model.fit(X_train=df.drop(columns=[target]),
                          y_train=df[target],
                          task='regression',
                          **args.get('using'))

            self.model_storage.json_set('args', args)
            self.model_storage.file_set('model', dill.dumps(model))

        else:
            raise Exception(
                "Data is empty!!"
            )

    def predict(self, df: pd.DataFrame, args: Optional[Dict] = None) -> pd.DataFrame:

        model = dill.loads(self.model_storage.file_get("model"))
        target = self.model_storage.json_get('args').get("target")

        results = pd.DataFrame(model.predict(df), columns=[target])

        return results



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    spreadsheet_id={
        'type': ARG_TYPE.STR,
        'description': 'The unique ID of the Google Sheet.'
    },
    sheet_name={
        'type': ARG_TYPE.STR,
        'description': 'The name of the sheet within the Google Sheet.'
    }
)

connection_args_example = OrderedDict(
    spreadsheet_id='12wgS-1KJ9ymUM-6VYzQ0nJYGitONxay7cMKLnEE2_d0',
    sheet_name='iris'
)



import unittest
from mindsdb.integrations.handlers.tdengine_handler.tdengine_handler import TDEngineHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class TDEngineHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "token": '<token_for_cloud>',
            "url": "********.cloud.tdengine.com",
            "database": "temp"
        }
        cls.handler = TDEngineHandler('test_tdengine_handler', connection_data=cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_connect(self):
        assert self.handler.connect()

    def test_2_create_table(self):
        query = "CREATE Table `hari` USING `temp` (`id`) TAGS (0);"
        result = self.handler.query(query)
        assert result.type is not RESPONSE_TYPE.ERROR
        pass

    def test_3_insert(self):
        query = "INSERT INTO hari  VALUES (NOW, 12);"
        result = self.handler.query(query)
        assert result.type is not RESPONSE_TYPE.ERROR

    def test_4_native_query_select(self):
        query = "SELECT * FROM hari;"
        result = self.handler.query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_5_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is RESPONSE_TYPE.TABLE

    def test_6_get_columns(self):
        columns = self.handler.get_columns('hari')
        assert columns.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



from typing import Optional
import pandas as pd

from sqlalchemy import create_engine

from mindsdb_sql_parser import parse_sql
from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)


logger = log.getLogger(__name__)


class SolrHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Solr SQL statements.
    """

    name = 'solr'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'solr'

        if ('host' not in connection_data) or ('port' not in connection_data) or ('collection' not in connection_data):
            raise Exception("The host, port and collection parameter should be provided!")

        optional_parameters = ['use_ssl', 'username', 'password']
        for parameter in optional_parameters:
            if parameter not in connection_data:
                connection_data[parameter] = None

        if connection_data.get('use_ssl', False):
            connection_data['use_ssl'] = True
        else:
            connection_data['use_ssl'] = False

        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """
        if self.is_connected is True:
            return self.connection

        config = {
            'username': self.connection_data.get('username'),
            'password': self.connection_data.get('password'),
            'host': self.connection_data.get('host'),
            'port': self.connection_data.get('port'),
            'server_path': self.connection_data.get('server_path', 'solr'),
            'collection': self.connection_data.get('collection'),
            'use_ssl': self.connection_data.get('use_ssl')
        }

        connection = create_engine("solr://{username}:{password}@{host}:{port}/{server_path}/{collection}/sql?use_ssl={use_ssl}".format(**config))
        self.is_connected = True
        self.connection = connection.connect()
        return self.connection

    def disconnect(self):
        """
        Close any existing connections.
        """
        if self.is_connected is False:
            return
        self.connection.close()
        self.is_connected = False
        return

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the Solr database
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Solr {self.connection_data["host"]}, {e}!')
            response.error_message = str(e)

        if response.success is True and need_to_close:
            self.disconnect()
        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        connection = self.connect()

        try:
            result = connection.execute(query)
            columns = list(result.keys())
            if result:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    pd.DataFrame(
                        result,
                        columns=columns
                    )
                )
            else:
                response = Response(RESPONSE_TYPE.OK)

        except Exception as e:
            logger.error(f'Error running query: {query} on {self.connection_data["host"]}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement.
        """
        return self.native_query(query.to_string())

    def get_tables(self) -> Response:
        """
        Get a list with all of the tables in Solr
        """
        result = {}
        result['data_frame'] = pd.DataFrame([self.connection_data.get('collection')])
        df = result.data_frame
        result.data_frame = df.rename(columns={df.columns[0]: 'table_name'})
        return result

    def get_columns(self, table_name) -> Response:
        """
        Show details about the table
        """
        q = f"select * from {table_name} limit 1"
        result = self.native_query(q)
        df = pd.DataFrame([[col] for col in result.data_frame.columns])
        result.data_frame = df.rename(columns={df.columns[0]: 'column_name'})
        return result



import re
import datetime as dt
from dateutil.relativedelta import relativedelta
from typing import List

import sqlalchemy as sa

from mindsdb_sql_parser import parse_sql, ParsingException
from mindsdb_sql_parser.ast.mindsdb import CreateJob
from mindsdb_sql_parser.ast import Select, Star, Identifier, BinaryOperation, Constant

from mindsdb.utilities.config import config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.exception import EntityNotExistsError, EntityExistsError
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.query_context.context_controller import query_context_controller
from mindsdb.interfaces.database.log import LogDBController

from mindsdb.utilities import log

logger = log.getLogger(__name__)

default_project = config.get("default_project")


def split_sql(sql):
    # split sql by ';' ignoring delimiter in quotes
    pattern = re.compile(r"""((?:[^;"']|"[^"]*"|'[^']*')+)""")
    return pattern.split(sql)[1::2]


def calc_next_date(schedule_str, base_date: dt.datetime):
    schedule_str = schedule_str.lower().strip()

    repeat_prefix = "every "
    if schedule_str.startswith(repeat_prefix):
        repeat_str = schedule_str[len(repeat_prefix) :]
    else:
        # TODO cron format
        raise NotImplementedError(f"Schedule: {schedule_str}")

    items = repeat_str.split()

    if len(items) == 1:
        value = "1"
        period = items[0]
    elif len(items) == 2:
        value, period = items
    else:
        raise Exception(f"Can't parse repeat string: {repeat_str}")

    if not value.isdigit():
        raise Exception(f"Number expected: {value}")
    value = int(value)
    if period in ("minute", "minutes", "min"):
        delta = dt.timedelta(minutes=value)
    elif period in ("hour", "hours"):
        delta = dt.timedelta(hours=value)
    elif period in ("day", "days"):
        delta = dt.timedelta(days=value)
    elif period in ("week", "weeks"):
        delta = dt.timedelta(days=value * 7)  # 1 week = 7 days
    elif period in ("month", "months"):
        delta = relativedelta(months=value)
    else:
        raise Exception(f"Unknown period: {period}")

    # period limitation disabled for now
    # config = Config()
    # is_cloud = config.get('cloud', False)
    # if is_cloud and ctx.user_class == 0:
    #     if delta < dt.timedelta(days=1):
    #         raise Exception("Minimal allowed period can't be less than one day")

    next_date = base_date + delta

    return next_date


def parse_job_date(date_str: str) -> dt.datetime:
    """
    Convert string used as job data to datetime object
    :param date_str:
    :return:
    """

    if date_str.upper() == "NOW":
        return dt.datetime.now()

    date_formats = ["%Y-%m-%d %H:%M:%S", "%Y-%m-%d"]
    date = None
    for date_format in date_formats:
        try:
            date = dt.datetime.strptime(date_str, date_format)
        except ValueError:
            pass
    if date is None:
        raise ValueError(f"Can't parse date: {date_str}")
    return date


class JobsController:
    def add(
        self,
        name: str,
        project_name: str,
        query: str,
        start_at: dt.datetime = None,
        end_at: dt.datetime = None,
        if_query: str = None,
        schedule_str: str = None,
    ) -> str:
        """
        Create a new job

        More info: https://docs.mindsdb.com/mindsdb_sql/sql/create/jobs#create-job

        :param name: name of the job
        :param project_name: project name
        :param query: sql query for job to execute, it could be several queries seperated by ';'
        :param start_at: datetime of first execution of the job, optional
        :param end_at: datetime after which job should not be executed anymore
        :param if_query: condition for job,
           if this query (or last from list of queries separated by ';') returns data and no error in queries:
              job will not be executed
        :param schedule_str: description how to repeat job
            at the moment supports: 'every <number> <dimension>' or 'every <dimension>'
        :return: name of created job
        """
        if not name.islower():
            raise ValueError(f"The name must be in lower case: {name}")

        project_controller = ProjectController()
        project = project_controller.get(name=project_name)

        # check if exists
        if self.get(name, project_name) is not None:
            raise EntityExistsError("Job already exists", name)

        if start_at is None:
            start_at = dt.datetime.now()

        if end_at is not None and end_at < start_at:
            raise Exception(f"Wrong end date {start_at} > {end_at}")

        # check sql = try to parse it
        for sql in split_sql(query):
            try:
                # replace template variables with null
                sql = re.sub(r"\{\{[\w\d]+}}", "", sql)

                parse_sql(sql)
            except ParsingException as e:
                raise ParsingException(f"Unable to parse: {sql}: {e}")

        if if_query is not None:
            for sql in split_sql(if_query):
                try:
                    # replace template variables with null
                    sql = re.sub(r"\{\{[\w\d]+}}", "", sql)

                    parse_sql(sql)
                except ParsingException as e:
                    raise ParsingException(f"Unable to parse: {sql}: {e}")

        # plan next run
        next_run_at = start_at

        if schedule_str is not None:
            # try to calculate schedule string
            calc_next_date(schedule_str, start_at)
        else:
            # no schedule for job end_at is meaningless
            end_at = None

        name = name.lower()

        # create job record
        record = db.Jobs(
            company_id=ctx.company_id,
            user_class=ctx.user_class,
            name=name,
            project_id=project.id,
            query_str=query,
            if_query_str=if_query,
            start_at=start_at,
            end_at=end_at,
            next_run_at=next_run_at,
            schedule_str=schedule_str,
        )
        db.session.add(record)
        db.session.commit()

        return name

    def create(self, name: str, project_name: str, query: CreateJob) -> str:
        """
        Create job using AST query
        :param name: name of the job
        :param project_name: project name
        :param query: AST query with job parameters
        :return: name of created job
        """

        if project_name is None:
            project_name = default_project

        start_at = None
        if query.start_str is not None:
            start_at = parse_job_date(query.start_str)
            if start_at < dt.datetime.now():
                start_at = dt.datetime.now()

        end_at = None
        if query.end_str is not None:
            end_at = parse_job_date(query.end_str)

        query_str = query.query_str
        if_query_str = query.if_query_str

        schedule_str = None
        if query.repeat_str is not None:
            schedule_str = "every " + query.repeat_str

        return self.add(
            name,
            project_name,
            query=query_str,
            start_at=start_at,
            end_at=end_at,
            if_query=if_query_str,
            schedule_str=schedule_str,
        )

    def delete(self, name, project_name):
        project_controller = ProjectController()
        project = project_controller.get(name=project_name)

        # check if exists
        record = (
            db.session.query(db.Jobs)
            .filter_by(company_id=ctx.company_id, name=name, project_id=project.id, deleted_at=sa.null())
            .first()
        )
        if record is None:
            raise EntityNotExistsError("Job does not exist", name)

        self._delete_record(record)
        db.session.commit()

        # delete context
        query_context_controller.drop_query_context("job", record.id)
        query_context_controller.drop_query_context("job-if", record.id)

    def _delete_record(self, record):
        record.deleted_at = dt.datetime.now()

    def get_list(self, project_name=None):
        query = db.session.query(db.Jobs).filter_by(company_id=ctx.company_id, deleted_at=sa.null())

        project_controller = ProjectController()
        if project_name is not None:
            project = project_controller.get(name=project_name)
            query = query.filter_by(project_id=project.id)

        data = []
        project_names = {i.id: i.name for i in project_controller.get_list()}
        for record in query:
            data.append(
                {
                    "id": record.id,
                    "name": record.name,
                    "project": project_names[record.project_id],
                    "start_at": record.start_at,
                    "end_at": record.end_at,
                    "next_run_at": record.next_run_at,
                    "schedule_str": record.schedule_str,
                    "query": record.query_str,
                    "if_query": record.if_query_str,
                    "variables": query_context_controller.get_context_vars("job", record.id),
                }
            )
        return data

    def get(self, name: str, project_name: str) -> dict:
        """
        Get info about job
        :param name: name of the job
        :param project_name: job's project
        :return: dict with info about job
        """

        project_controller = ProjectController()
        project = project_controller.get(name=project_name)

        record = (
            db.session.query(db.Jobs)
            .filter_by(company_id=ctx.company_id, name=name, project_id=project.id, deleted_at=sa.null())
            .first()
        )

        if record is not None:
            return {
                "id": record.id,
                "name": record.name,
                "project": project_name,
                "start_at": record.start_at,
                "end_at": record.end_at,
                "next_run_at": record.next_run_at,
                "schedule_str": record.schedule_str,
                "query": record.query_str,
                "if_query": record.if_query_str,
                "variables": query_context_controller.get_context_vars("job", record.id),
            }

    def get_history(self, name: str, project_name: str) -> List[dict]:
        """
        Get history of the job's calls
        :param name: job name
        :param project_name: project name
        :return: List of job executions
        """

        logs_db_controller = LogDBController()

        query = Select(
            targets=[Star()],
            from_table=Identifier("jobs_history"),
            where=BinaryOperation(
                op="and",
                args=[
                    BinaryOperation(op="=", args=[Identifier("name"), Constant(name)]),
                    BinaryOperation(op="=", args=[Identifier("project"), Constant(project_name)]),
                ],
            ),
        )
        response = logs_db_controller.query(query)

        names = [i["name"] for i in response.columns]
        return response.data_frame[names].to_dict(orient="records")


class JobsExecutor:
    def get_next_tasks(self):
        # filter next_run < now
        query = (
            db.session.query(db.Jobs)
            .filter(
                db.Jobs.next_run_at < dt.datetime.now(),
                db.Jobs.deleted_at == sa.null(),
                db.Jobs.active == True,  # noqa
            )
            .order_by(db.Jobs.next_run_at)
        )

        return query.all()

    def update_task_schedule(self, record):
        # calculate next run

        if record.next_run_at > dt.datetime.now():
            # do nothing, it is already planned in future
            return

        if record.schedule_str is None:
            # not need to run it anymore
            self._delete_record(record)
            return

        next_run_at = calc_next_date(record.schedule_str, base_date=record.next_run_at)

        if next_run_at is None:
            # no need to run it
            self._delete_record(record)
        elif record.end_at is not None and next_run_at > record.end_at:
            self._delete_record(record)
        else:
            # plan next run, but not in the past
            if next_run_at < dt.datetime.now():
                next_run_at = dt.datetime.now()
            record.next_run_at = next_run_at

    def _delete_record(self, record):
        record.deleted_at = dt.datetime.now()

    def lock_record(self, record_id):
        # workaround for several concurrent workers on cloud:
        #  create history record before start of task
        record = db.Jobs.query.get(record_id)

        try:
            history_record = db.JobsHistory(job_id=record.id, start_at=record.next_run_at, company_id=record.company_id)

            db.session.add(history_record)
            db.session.commit()

            return history_record.id

        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            db.session.rollback()

            # check if it is an old lock
            history_record = db.JobsHistory.query.filter_by(
                job_id=record.id, start_at=record.next_run_at, company_id=record.company_id
            ).first()
            if history_record.updated_at < dt.datetime.now() - dt.timedelta(seconds=30):
                db.session.delete(history_record)
                db.session.commit()

        return None

    def __fill_variables(self, sql, record, history_record):
        if "{{PREVIOUS_START_DATETIME}}" in sql:
            # get previous run date
            history_prev = (
                db.session.query(db.JobsHistory.start_at)
                .filter(db.JobsHistory.job_id == record.id, db.JobsHistory.id != history_record.id)
                .order_by(db.JobsHistory.id.desc())
                .first()
            )
            if history_prev is None:
                # start date of the job
                value = record.created_at
            else:
                # fix for twitter: created_at filter must be minimum of 10 seconds prior to the current time
                value = history_prev.start_at - dt.timedelta(seconds=60)
            value = value.strftime("%Y-%m-%d %H:%M:%S")
            sql = sql.replace("{{PREVIOUS_START_DATETIME}}", value)

        if "{{START_DATE}}" in sql:
            value = history_record.start_at.strftime("%Y-%m-%d")
            sql = sql.replace("{{START_DATE}}", value)
        if "{{START_DATETIME}}" in sql:
            value = history_record.start_at.strftime("%Y-%m-%d %H:%M:%S")
            sql = sql.replace("{{START_DATETIME}}", value)
        return sql

    def execute_task_local(self, record_id, history_id=None):
        record = db.Jobs.query.get(record_id)

        # set up environment

        ctx.set_default()
        ctx.company_id = record.company_id
        if record.user_class is not None:
            ctx.user_class = record.user_class

        if history_id is None:
            history_record = db.JobsHistory(
                job_id=record.id,
                start_at=dt.datetime.now(),
                company_id=record.company_id,
            )
            db.session.add(history_record)
            db.session.flush()
            history_id = history_record.id
            db.session.commit()

        else:
            history_record = db.JobsHistory.query.get(history_id)

        project_controller = ProjectController()
        project = project_controller.get(record.project_id)
        executed_sql = ""

        from mindsdb.api.executor.controllers.session_controller import SessionController
        from mindsdb.api.executor.command_executor import ExecuteCommands

        sql_session = SessionController()
        sql_session.database = project.name
        command_executor = ExecuteCommands(sql_session)

        # job with condition?
        query_context_controller.set_context("job-if", record.id)
        error = ""
        to_execute_query = True
        if record.if_query_str is not None:
            data = None
            for sql in split_sql(record.if_query_str):
                try:
                    #  fill template variables
                    sql = self.__fill_variables(sql, record, history_record)

                    query = parse_sql(sql)
                    executed_sql += sql + "; "

                    ret = command_executor.execute_command(query)
                    if ret.error_code is not None:
                        error = ret.error_message
                        break

                    data = ret.data
                except Exception as e:
                    logger.error(e)
                    error = str(e)
                    break

            # check error or last result
            if error or data is None or len(data) == 0:
                to_execute_query = False

        query_context_controller.release_context("job-if", record.id)
        if to_execute_query:
            query_context_controller.set_context("job", record.id)
            for sql in split_sql(record.query_str):
                try:
                    #  fill template variables
                    sql = self.__fill_variables(sql, record, history_record)

                    query = parse_sql(sql)
                    executed_sql += sql + "; "

                    ret = command_executor.execute_command(query)
                    if ret.error_code is not None:
                        error = ret.error_message
                        break
                except Exception as e:
                    logger.error(e)
                    error = str(e)
                    break

        try:
            self.update_task_schedule(record)
        except Exception as e:
            db.session.rollback()
            logger.error(f"Error to update schedule: {e}")
            error += f"Error to update schedule: {e}"

            # stop scheduling
            record.next_run_at = None

        history_record = db.JobsHistory.query.get(history_id)

        if error:
            history_record.error = error
        history_record.end_at = dt.datetime.now()
        history_record.query_str = executed_sql

        db.session.commit()



import sys
import argparse
import asyncio
from typing import List, Dict
from contextlib import AsyncExitStack

from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client

from mindsdb.utilities import log
from mindsdb.interfaces.agents.mcp_client_agent import create_mcp_agent

logger = log.getLogger(__name__)


async def run_conversation(agent_wrapper, messages: List[Dict[str, str]], stream: bool = False):
    """Run a conversation with the agent and print responses"""
    try:
        if stream:
            logger.info("Streaming response:")
            async for chunk in agent_wrapper.acompletion_stream(messages):
                content = chunk["choices"][0]["delta"].get("content", "")
                if content:
                    # We still need to print content for streaming display
                    # but we'll log it as debug as well
                    logger.debug(f"Stream content: {content}")
                    sys.stdout.write(content)
                    sys.stdout.flush()
            logger.debug("End of stream")
            sys.stdout.write("\n\n")
            sys.stdout.flush()
        else:
            logger.info("Getting response...")
            response = await agent_wrapper.acompletion(messages)
            content = response["choices"][0]["message"]["content"]
            logger.info(f"Response: {content}")
            # We still need to display the response to the user
            sys.stdout.write(f"{content}\n")
            sys.stdout.flush()
    except Exception as e:
        logger.error(f"Error during agent conversation: {str(e)}")


async def execute_direct_query(query):
    """Execute a direct SQL query using MCP"""
    logger.info(f"Executing direct SQL query: {query}")

    # Set up MCP client to connect to the running server
    async with AsyncExitStack() as stack:
        # Connect to MCP server
        server_params = StdioServerParameters(
            command="python",
            args=["-m", "mindsdb", "--api=mcp"],
            env=None
        )

        try:
            stdio_transport = await stack.enter_async_context(stdio_client(server_params))
            stdio, write = stdio_transport
            session = await stack.enter_async_context(ClientSession(stdio, write))

            await session.initialize()

            # List available tools
            tools_response = await session.list_tools()
            tool_names = [tool.name for tool in tools_response.tools]
            logger.info(f"Available tools: {tool_names}")

            # Find query tool
            query_tool = None
            for tool in tools_response.tools:
                if tool.name == "query":
                    query_tool = tool
                    break

            if not query_tool:
                logger.error("No 'query' tool found in MCP server")
                return

            # Execute query
            result = await session.call_tool("query", {"query": query})
            logger.info(f"Query result: {result.content}")
        except Exception as e:
            logger.error(f"Error executing query: {str(e)}")
            logger.info("Make sure the MindsDB server is running with MCP enabled: python -m mindsdb --api=mysql,mcp,http")


async def main():
    parser = argparse.ArgumentParser(description="Run an agent as an MCP client")
    parser.add_argument("--agent", type=str, help="Name of the agent to use")
    parser.add_argument("--project", type=str, default="mindsdb", help="Project containing the agent")
    parser.add_argument("--host", type=str, default="127.0.0.1", help="MCP server host")
    parser.add_argument("--port", type=int, default=47337, help="MCP server port")
    parser.add_argument("--query", type=str, help="Query to send to the agent")
    parser.add_argument("--stream", action="store_true", help="Stream the response")
    parser.add_argument("--execute-direct", type=str, help="Execute a direct SQL query via MCP (for testing)")

    args = parser.parse_args()

    try:
        # Initialize database connection
        from mindsdb.interfaces.storage import db
        db.init()

        # Direct SQL execution mode (for testing MCP connection)
        if args.execute_direct:
            await execute_direct_query(args.execute_direct)
            return 0

        # Make sure agent name is provided
        if not args.agent:
            parser.error("the --agent argument is required unless --execute-direct is used")

        # Create the agent
        logger.info(f"Creating MCP client agent for '{args.agent}' in project '{args.project}'")
        logger.info(f"Connecting to MCP server at {args.host}:{args.port}")
        logger.info("Make sure MindsDB server is running with MCP enabled: python -m mindsdb --api=mysql,mcp,http")

        agent_wrapper = create_mcp_agent(
            agent_name=args.agent,
            project_name=args.project,
            mcp_host=args.host,
            mcp_port=args.port
        )

        # Run an example query if provided
        if args.query:
            messages = [{"role": "user", "content": args.query}]
            await run_conversation(agent_wrapper, messages, args.stream)
        else:
            # Interactive mode
            logger.info("Entering interactive mode. Type 'exit' to quit.")
            logger.info("Available commands: exit/quit, clear, sql:")

            # We still need to show these instructions to the user
            sys.stdout.write("\nEntering interactive mode. Type 'exit' to quit.\n")
            sys.stdout.write("\nAvailable commands:\n")
            sys.stdout.write("  exit, quit - Exit the program\n")
            sys.stdout.write("  clear - Clear conversation history\n")
            sys.stdout.write("  sql: <query> - Execute a direct SQL query via MCP\n")
            sys.stdout.flush()

            messages = []

            while True:
                # We need to keep input for user interaction
                user_input = input("\nYou: ")

                # Check for special commands
                if user_input.lower() in ["exit", "quit"]:
                    logger.info("Exiting interactive mode")
                    break
                elif user_input.lower() == "clear":
                    messages = []
                    logger.info("Conversation history cleared")
                    sys.stdout.write("Conversation history cleared\n")
                    sys.stdout.flush()
                    continue
                elif user_input.lower().startswith("sql:"):
                    # Direct SQL execution using the agent's session
                    sql_query = user_input[4:].strip()
                    logger.info(f"Executing SQL: {sql_query}")
                    try:
                        # Use the tool from the agent
                        if hasattr(agent_wrapper.agent, "session") and agent_wrapper.agent.session:
                            result = await agent_wrapper.agent.session.call_tool("query", {"query": sql_query})
                            logger.info(f"SQL result: {result.content}")
                            # We need to show the result to the user
                            sys.stdout.write(f"Result: {result.content}\n")
                            sys.stdout.flush()
                        else:
                            logger.error("No active MCP session")
                            sys.stdout.write("Error: No active MCP session\n")
                            sys.stdout.flush()
                    except Exception as e:
                        logger.error(f"SQL Error: {str(e)}")
                        sys.stdout.write(f"SQL Error: {str(e)}\n")
                        sys.stdout.flush()
                    continue

                messages.append({"role": "user", "content": user_input})
                await run_conversation(agent_wrapper, messages, args.stream)

                # Add assistant's response to the conversation history
                if not args.stream:
                    response = await agent_wrapper.acompletion(messages)
                    messages.append({
                        "role": "assistant",
                        "content": response["choices"][0]["message"]["content"]
                    })

        # Clean up resources
        logger.info("Cleaning up resources")
        await agent_wrapper.cleanup()

    except Exception as e:
        logger.error(f"Error running MCP agent: {str(e)}")
        logger.info("Make sure the MindsDB server is running with MCP enabled: python -m mindsdb --api=mysql,mcp,http")
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(asyncio.run(main()))



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    address={
        'type': ARG_TYPE.STR,
        'description': 'The hostname, IP address, or URL of the SAP HANA database.',
        'required': True,
        'label': 'Address'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The port number for connecting to the SAP HANA database.',
        'required': True,
        'label': 'Port'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The username for the SAP HANA database.',
        'required': True,
        'label': 'User'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password for the SAP HANA database.',
        'secret': True,
        'required': True,
        'label': 'Password'
    },
    schema={
        'type': ARG_TYPE.STR,
        'description': "The database schema to use. Defaults to the user's default schema.",
        'required': False,
        'label': 'Schema'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The name of the database to connect to. This parameter is not used for SAP HANA Cloud.',
        'required': False,
        'label': 'Database'
    },
    encrypt={
        'type': ARG_TYPE.BOOL,
        'description': 'The setting to enable or disable encryption. Default is `True`.',
        'required': False,
        'label': 'Encrypt'
    }
)

connection_args_example = OrderedDict(
    host='123e4567-e89b-12d3-a456-426614174000.hana.trial-us10.hanacloud.ondemand.com',
    port=30013,
    user='DBADMIN',
    password='password'
)



import os

# By default, we have Open Telemetry SDK enabled on all envs, except for local which is disabled by default.
OTEL_SDK_DISABLED = (os.getenv("OTEL_SDK_DISABLED", "false").lower() == "true"
                     or os.getenv("OTEL_SERVICE_ENVIRONMENT", "local").lower() == "local")

# If you want to enable Open Telemetry on local for some reason please set OTEL_SDK_FORCE_RUN to true
OTEL_SDK_FORCE_RUN = os.getenv("OTEL_SDK_FORCE_RUN", "false").lower() == "true"

OTEL_ENABLED = not OTEL_SDK_DISABLED or OTEL_SDK_FORCE_RUN

def increment_otel_query_request_counter(metadata: dict) -> None:
    pass

trace = None
if OTEL_ENABLED:
    try:
        from mindsdb.utilities.otel.prepare import trace
        from mindsdb.utilities.otel.metric_handlers import increment_otel_query_request_counter
    except Exception:
        pass




import pytest
import psycopg2
import os

from mindsdb.integrations.handlers.postgres_handler.postgres_handler import (
    PostgresHandler,
)
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE

HANDLER_KWARGS = {
    "connection_data": {
        "host": os.environ.get("MDB_TEST_POSTGRES_HOST", "127.0.0.1"),
        "port": os.environ.get("MDB_TEST_POSTGRES_PORT", "5432"),
        "user": os.environ.get("MDB_TEST_POSTGRES_USER", "postgres"),
        "password": os.environ.get("MDB_TEST_POSTGRES_PASSWORD", "supersecret"),
        "database": os.environ.get(
            "MDB_TEST_POSTGRES_DATABASE", "postgres_db_handler_test"
        ),
    }
}

table_for_creation = "test_mdb"
expected_columns = ["col_one", "col_two", "col_three", "col_four"]


curr_dir = os.path.dirname(os.path.realpath(__file__))


def seed_db():
    """Seed the test DB with some data"""

    conn_info = HANDLER_KWARGS["connection_data"].copy()
    conn_info["database"] = "postgres"
    db = psycopg2.connect(**conn_info)
    db.autocommit = True
    cursor = db.cursor()

    try:
        cursor.execute("DROP DATABASE IF EXISTS postgres_db_handler_test;")
        db.commit()

        # Create the test database
        cursor.execute("CREATE DATABASE postgres_db_handler_test;")

        # Reconnect to the new database
        conn_info["database"] = "postgres_db_handler_test"
        db = psycopg2.connect(**conn_info)
        db.autocommit = True
        cursor = db.cursor()

        # Seed the database with data
        with open(os.path.join(curr_dir, "seed.sql"), "r") as f:
            cursor.execute(f.read())
        db.commit()

    finally:
        # Close the cursor and the connection
        cursor.close()
        db.close()


@pytest.fixture(scope="module")
def handler(request):
    seed_db()
    handler = PostgresHandler(**HANDLER_KWARGS)
    yield handler


def check_valid_response(res):
    if res.resp_type == RESPONSE_TYPE.TABLE:
        assert res.data_frame is not None, "expected to have some data, but got None"
    assert (
        res.error_code == 0
    ), f"expected to have zero error_code, but got {res.error_code}"
    assert (
        res.error_message is None
    ), f"expected to have None in error message, but got {res.error_message}"


def get_table_names(handler):
    res = handler.get_tables()
    tables = res.data_frame
    assert tables is not None, "expected to have some tables in the db, but got None"
    assert (
        "table_name" in tables
    ), f"expected to get 'table_name' column in the response:\n{tables}"
    return list(tables["table_name"])


class TestPostgresConnection:
    def test_connect(self, handler):
        handler.connect()
        assert handler.is_connected, "connection error"

    def test_check_connection(self, handler):
        res = handler.check_connection()
        assert res.success, res.error_message


class TestPostgresQuery:
    def test_native_query_show_dbs(self, handler):
        dbs = handler.native_query("SELECT datname FROM pg_database")
        dbs = dbs.data_frame
        assert dbs is not None, "expected to get some data, but got None"
        assert "datname" in dbs, f"Expected to get 'datname' column in response:\n{dbs}"
        dbs = list(dbs["datname"])
        expected_db = HANDLER_KWARGS["connection_data"]["database"]
        assert (
            expected_db in dbs
        ), f"expected to have {expected_db} db in response: {dbs}"

    def test_select_query(self, handler):
        limit = 3
        query = "SELECT * FROM test"
        res = handler.query(query)
        check_valid_response(res)
        got_rows = res.data_frame.shape[0]
        want_rows = limit
        assert (
            got_rows == want_rows
        ), f"expected to have {want_rows} rows in response but got: {got_rows}"


class TestPostgresTables:
    def test_get_tables(self, handler):
        res = handler.get_tables()
        tables = res.data_frame
        test_table = get_table_names(handler)
        assert (
            tables is not None
        ), "expected to have some table_name in the db, but got None"
        assert (
            "table_name" in tables
        ), f"expected to get 'table_name' column in the response:\n{tables}"
        # get a specific table from the tables list
        assert (
            "test" in test_table
        ), f"expected to have 'test' table in the db but got: {test_table}"

    def test_create_table(self, handler):
        new_table = table_for_creation
        res = handler.native_query(
            f"CREATE TABLE IF NOT EXISTS {new_table} (test_col INT)"
        )
        check_valid_response(res)
        tables = get_table_names(handler)
        assert (
            new_table in tables
        ), f"expected to have {new_table} in database, but got: {tables}"

    def test_drop_table(self, handler):
        drop_table = table_for_creation
        res = handler.native_query(f"DROP TABLE IF EXISTS {drop_table}")
        check_valid_response(res)
        tables = get_table_names(handler)
        assert drop_table not in tables


class TestPostgresColumns:
    def test_get_columns(self, handler):
        described = handler.get_columns("test")
        describe_data = described.data_frame
        check_valid_response(described)
        got_columns = list(describe_data.iloc[:, 0])
        assert set(got_columns) == set(
            expected_columns
        ), f"expected to have next columns in test table:\n{expected_columns}\nbut got:\n{got_columns}"


class TestPostgresDisconnect:
    def test_disconnect(self, handler):
        handler.disconnect()
        assert handler.is_connected is False, "failed to disconnect"

    def test_check_connection(self, handler):
        res = handler.check_connection()
        assert res.success, res.error_message



import time
from typing import Optional, List

import duckdb
from typing import Any

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import RESPONSE_TYPE, HandlerResponse, HandlerStatusResponse
from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender


class DummyHandler(DatabaseHandler):
    name = "dummy_data"

    def __init__(self, **kwargs):
        super().__init__("dummy_data")
        self.db_path = None

        args = kwargs.get("connection_data", {})
        if "db_path" in args:
            self.db_path = args["db_path"]

    def connect(self):
        """Set up any connections required by the handler"""
        return self.db_path is not None

    def disconnect(self):
        """Close any existing connections"""
        return

    def check_connection(self) -> HandlerStatusResponse:
        """Check connection to the handler

        Returns:
            HandlerStatusResponse
        """
        return HandlerStatusResponse(success=True)

    def native_query(self, query: Any, params: Optional[List] = None) -> HandlerResponse:
        """Receive raw query and act upon it somehow

        Args:
            query (Any): query in native format (str for sql databases, etc)
            params (Optional[List])

        Returns:
            HandlerResponse
        """
        con = duckdb.connect(self.db_path)
        if params is not None:
            query = query.replace("%s", "?")
            cur = con.executemany(query, params)
            if cur.rowcount >= 0:
                result_df = cur.fetchdf()
            else:
                con.close()
                return HandlerResponse(RESPONSE_TYPE.OK)
        else:
            result_df = con.execute(query).fetchdf()
        con.close()
        return HandlerResponse(RESPONSE_TYPE.TABLE, result_df)

    def query(self, query: ASTNode) -> HandlerResponse:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INSERT, DELETE, etc

        Returns:
            HandlerResponse
        """
        renderer = SqlalchemyRender("postgres")
        query_str, params = renderer.get_exec_params(query, with_failback=True)
        return self.native_query(query_str, params)

    def get_tables(self) -> HandlerResponse:
        """Get a list of all the tables in the database

        Returns:
            HandlerResponse: Names of the tables in the database
        """
        q = "SHOW TABLES;"
        result = self.native_query(q)
        df = result.data_frame
        result.data_frame = df.rename(columns={df.columns[0]: "table_name"})
        return result

    def get_columns(self, table_name: str) -> HandlerResponse:
        """Get details about a table

        Args:
            table_name (str): Name of the table to retrieve details of.

        Returns:
            HandlerResponse: Details of the table.
        """
        query = f"DESCRIBE {table_name};"
        return self.native_query(query)

    def subscribe(self, stop_event, callback, table_name, columns=None, **kwargs):
        while True:
            if stop_event.is_set():
                return
            time.sleep(0.3)



import os
from typing import List, Iterator
from langchain_core.documents import Document as LangchainDocument
from langchain_text_splitters import MarkdownHeaderTextSplitter

from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.integrations.utilities.rag.loaders.file_loader import FileLoader
from mindsdb.integrations.utilities.rag.splitters.file_splitter import (
    FileSplitter,
)
from mindsdb.integrations.handlers.web_handler.urlcrawl_helpers import get_all_websites
from mindsdb.interfaces.knowledge_base.preprocessing.models import Document
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class DocumentLoader:
    """Handles loading documents from various sources including SQL queries"""

    def __init__(
        self,
        file_controller: FileController,
        file_splitter: FileSplitter,
        markdown_splitter: MarkdownHeaderTextSplitter,
        file_loader_class=FileLoader,
        mysql_proxy=None,
    ):
        """
        Initialize with required dependencies

        Args:
            file_controller: Controller for file operations
            file_splitter: Splitter for file content
            markdown_splitter: Splitter for markdown content
            file_loader_class: Class to use for file loading
            mysql_proxy: Proxy for executing MySQL queries
        """
        self.file_controller = file_controller
        self.file_splitter = file_splitter
        self.markdown_splitter = markdown_splitter
        self.file_loader_class = file_loader_class
        self.mysql_proxy = mysql_proxy

    def load_files(self, file_names: List[str]) -> Iterator[Document]:
        """Load and split documents from files"""
        for file_name in file_names:
            file_path = self.file_controller.get_file_path(file_name)
            loader = self.file_loader_class(file_path)

            for doc in loader.lazy_load():
                # Add file extension to metadata for proper splitting
                extension = os.path.splitext(file_path)[1].lower()
                doc.metadata["extension"] = extension
                doc.metadata["source"] = file_name

                # Use FileSplitter to handle the document based on its type
                split_docs = self.file_splitter.split_documents([doc])
                for split_doc in split_docs:
                    # Preserve original metadata while adding split-specific metadata
                    metadata = doc.metadata.copy()
                    metadata.update(split_doc.metadata or {})

                    yield Document(content=split_doc.page_content, metadata=metadata)

    def load_web_pages(
        self,
        urls: List[str],
        crawl_depth: int,
        limit: int,
        filters: List[str] = None,
    ) -> Iterator[Document]:
        """Load and split documents from web pages"""
        websites_df = get_all_websites(urls, crawl_depth=crawl_depth, limit=limit, filters=filters)

        for _, row in websites_df.iterrows():
            # Create a document with HTML extension for proper splitting
            doc = LangchainDocument(
                page_content=row["text_content"], metadata={"extension": ".html", "url": row["url"]}
            )

            # Use FileSplitter to handle HTML content
            split_docs = self.file_splitter.split_documents([doc])
            for split_doc in split_docs:
                metadata = doc.metadata.copy()
                metadata.update(split_doc.metadata or {})

                yield Document(content=split_doc.page_content, metadata=metadata)



import pandas as pd
from typing import Text, List, Dict, Any

from mindsdb_sql_parser import ast

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions

from mindsdb.integrations.libs.response import HandlerResponse as Response

from mindsdb.integrations.utilities.handlers.query_utilities.insert_query_utilities import (
    INSERTQueryParser
)

from mindsdb.utilities import log

logger = log.getLogger(__name__)


class MessagesTable(APITable):
    """The Discord Table implementation"""

    def get_columns(self):
        return [
            'id',
            'type',
            'content',
            'author_id',
            'author_username',
            'author_global_name',
            'author_avatar',
            'author_banner_color',
            'attachments',
            'embeds',
            'mentions',
            'mention_roles',
            'pinned',
            'mention_everyone',
            'tts',
            'timestamp',
            'edited_timestamp',
            'flags',
            'components',
            'nonce',
            'referenced_message',
        ]

    def select(self, query: ast.Select) -> Response:
        """Selects data from the Discord channel.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query.

        Returns
        -------
        Response
            Response object representing collected data from Discord.
        """

        conditions = extract_comparison_conditions(query.where)

        params = {}
        filters = []
        user_filter_flag = channel_filter_flag = False
        for op, arg1, arg2 in conditions:
            if op == 'or':
                raise NotImplementedError('OR is not supported')
            if arg1 == 'timestamp':
                if op == '>':
                    params['after'] = arg2
                elif op == '<':
                    params['before'] = arg2
                else:
                    raise NotImplementedError(
                        f'Unsupported operator {op} for timestamp'
                    )

            elif arg1 in ['author_id', 'author_username', 'author_global_name']:
                if user_filter_flag:
                    raise NotImplementedError('Multiple user filters are not supported')
                user_filter_flag = True

                if op != '=':
                    raise NotImplementedError(f'Unsupported operator {op} for {arg1}')

                # if arg1 == 'author_id':
                #     filters.append(lambda x: x.author.id == int(arg2))
                # elif arg1 == 'author_username':
                #     filters.append(lambda x: x.author.username == arg2)
                # elif arg1 == 'author_global_name':
                #     filters.append(lambda x: x.author.global_name == arg2)

            elif arg1 == 'channel_id':
                if op != '=':
                    raise NotImplementedError(f'Unsupported operator {op} for {arg1}')
                channel_filter_flag = True
                params['channel_id'] = int(arg2)

            else:
                filters.append([op, arg1, arg2])

        if query.limit is not None:
            params['limit'] = query.limit
        else:
            params['limit'] = 100

        if not channel_filter_flag:
            raise NotImplementedError('Channel filter is required')

        result = self.handler.call_discord_api(
            'get_messages', params=params, filters=filters
        )

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]
        return result

    def insert(self, query: ast.Insert) -> None:
        """Sends messages to a Discord channel.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['channel_id', 'text'],
            mandatory_columns=['channel_id', 'text'],
            all_mandatory=False,
        )
        message_data = insert_statement_parser.parse_query()
        self.send_message(message_data)

    def send_message(self, message_data: List[Dict[Text, Any]]) -> None:
        """Sends messages to a Discord Channel using the parsed message data.

        Parameters
        ----------
        message_data : List[Dict[Text, Any]]
           List of dictionaries containing the messages to send.

        Returns
        -------
        None
        """
        for message in message_data:
            try:
                params = {'channel_id': message['channel_id'], 'text': message['text']}
                self.handler.call_discord_api('send_message', params=params)
                logger.info("Message sent to Discord channel successfully.")
            except Exception as e:
                logger.error(f"Error sending message to Discord channel: {e}")
                raise e



import shopify
import requests

from mindsdb.integrations.handlers.shopify_handler.shopify_tables import ProductsTable, CustomersTable, OrdersTable, InventoryLevelTable, LocationTable, CustomerReviews, CarrierServiceTable, ShippingZoneTable, SalesChannelTable
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)

from mindsdb.utilities import log
from mindsdb_sql_parser import parse_sql
from mindsdb.integrations.libs.api_handler_exceptions import InvalidNativeQuery, ConnectionFailed, MissingConnectionParams

logger = log.getLogger(__name__)


class ShopifyHandler(APIHandler):
    """
    The Shopify handler implementation.
    """

    name = 'shopify'

    def __init__(self, name: str, **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        if kwargs.get("connection_data") is None:
            raise MissingConnectionParams("Incomplete parameters passed to Shopify Handler")

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        products_data = ProductsTable(self)
        self._register_table("products", products_data)

        customers_data = CustomersTable(self)
        self._register_table("customers", customers_data)

        orders_data = OrdersTable(self)
        self._register_table("orders", orders_data)

        inventory_level_data = InventoryLevelTable(self)
        self._register_table("inventory_level", inventory_level_data)

        location_data = LocationTable(self)
        self._register_table("locations", location_data)

        customer_reviews_data = CustomerReviews(self)
        self._register_table("customer_reviews", customer_reviews_data)

        carrier_service_data = CarrierServiceTable(self)
        self._register_table("carrier_service", carrier_service_data)

        shipping_zone_data = ShippingZoneTable(self)
        self._register_table("shipping_zone", shipping_zone_data)

        sales_channel_data = SalesChannelTable(self)
        self._register_table("sales_channel", sales_channel_data)

    def connect(self):
        """
        Set up the connection required by the handler.
        Returns
        -------
        StatusResponse
            connection object
        """
        if self.is_connected is True:
            return self.connection

        if self.kwargs.get("connection_data") is None:
            raise MissingConnectionParams("Incomplete parameters passed to Shopify Handler")

        api_session = shopify.Session(self.connection_data['shop_url'], '2021-10', self.connection_data['access_token'])

        self.yotpo_app_key = self.connection_data['yotpo_app_key'] if 'yotpo_app_key' in self.connection_data else None
        self.yotpo_access_token = self.connection_data['yotpo_access_token'] if 'yotpo_access_token' in self.connection_data else None

        self.connection = api_session

        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)

        try:
            api_session = self.connect()
            shopify.ShopifyResource.activate_session(api_session)
            shopify.Shop.current()
            response.success = True
        except Exception as e:
            logger.error('Error connecting to Shopify!')
            raise ConnectionFailed("Conenction to Shopify failed.")
            response.error_message = str(e)

        if self.yotpo_app_key is not None and self.yotpo_access_token is not None:
            url = f"https://api.yotpo.com/v1/apps/{self.yotpo_app_key}/reviews?count=1&utoken={self.yotpo_access_token}"
            headers = {
                "accept": "application/json",
                "Content-Type": "application/json"
            }
            if requests.get(url, headers=headers).status_code == 200:
                response.success = True
            else:
                response.success = False

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        try:
            ast = parse_sql(query)
        except Exception:
            raise InvalidNativeQuery(f"The query {query} is invalid.")
        return self.query(ast)



from typing import Dict, Text, Callable, Union

from botbuilder.schema import Activity, ActivityTypes
from botbuilder.schema import ChannelAccount
from botframework.connector import ConnectorClient
from botframework.connector.auth import MicrosoftAppCredentials
import msal
from requests.exceptions import RequestException

from mindsdb.integrations.handlers.ms_teams_handler.ms_graph_api_teams_client import (
    MSGraphAPIBaseClient,
    MSGraphAPITeamsApplicationPermissionsClient,
    MSGraphAPITeamsDelegatedPermissionsClient
)
from mindsdb.integrations.handlers.ms_teams_handler.ms_teams_tables import (
    ChannelsTable, ChannelMessagesTable, ChatsTable, ChatMessagesTable, TeamsTable
)
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.integrations.libs.api_handler import APIChatHandler
from mindsdb.integrations.utilities.handlers.auth_utilities.microsoft import (
    MSGraphAPIApplicationPermissionsManager,
    MSGraphAPIDelegatedPermissionsManager
)
from mindsdb.integrations.utilities.handlers.auth_utilities.exceptions import AuthException
from mindsdb.interfaces.chatbot.types import ChatBotMessage
from mindsdb.utilities import log


logger = log.getLogger(__name__)


def chatbot_only(func):
    def wrapper(self, *args, **kwargs):
        if self.connection_data.get('opertion_mode', 'datasource') != 'chatbot':
            raise ValueError("This connection can only be used as a data source. Please use a chatbot connection by setting the 'mode' parameter to 'chat'.")
        return func(self, *args, **kwargs)
    return wrapper


class MSTeamsHandler(APIChatHandler):
    """
    This handler handles the connection and execution of SQL statements on Microsoft Teams via the Microsoft Graph API.
    It is also responsible for handling the chatbot functionality.
    """

    name = 'teams'

    def __init__(self, name: str, **kwargs):
        """
        Initializes the handler.

        Args:
            name (str): name of particular handler instance
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.handler_storage = kwargs['handler_storage']
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        self.service_url = None
        self.channel_id = None
        self.bot_id = None
        self.conversation_id = None

    def connect(self) -> Union[MicrosoftAppCredentials, MSGraphAPIBaseClient]:
        """
        Establishes a connection to the Microsoft Teams registered app or the Microsoft Graph API.

        Returns:
            Union[MicrosoftAppCredentials, MSGraphAPITeamsDelegatedPermissionsClient]: A connection object to the Microsoft Teams registered app or the Microsoft Graph API.
        """
        if self.is_connected:
            return self.connection

        # The default operation mode is 'datasource'. This is used for data source connections.
        operation_mode = self.connection_data.get('operation_mode', 'datasource')
        if operation_mode == 'datasource':
            # Initialize the token cache.
            cache = msal.SerializableTokenCache()

            # Load the cache from file if it exists.
            cache_file = 'cache.bin'
            try:
                cache_content = self.handler_storage.file_get(cache_file)
            except FileNotFoundError:
                cache_content = None

            if cache_content:
                cache.deserialize(cache_content)

            # The default permissions mode is 'delegated'. This requires the user to sign in.
            permission_mode = self.connection_data.get('permission_mode', 'delegated')
            if permission_mode == 'delegated':
                permissions_manager = MSGraphAPIDelegatedPermissionsManager(
                    client_id=self.connection_data['client_id'],
                    client_secret=self.connection_data['client_secret'],
                    tenant_id=self.connection_data['tenant_id'],
                    cache=cache,
                    code=self.connection_data.get('code')
                )

            elif permission_mode == 'application':
                permissions_manager = MSGraphAPIApplicationPermissionsManager(
                    client_id=self.connection_data['client_id'],
                    client_secret=self.connection_data['client_secret'],
                    tenant_id=self.connection_data['tenant_id'],
                    cache=cache
                )

            else:
                raise ValueError("The supported permission modes are 'delegated' and 'application'.")

            access_token = permissions_manager.get_access_token()

            # Save the cache back to file if it has changed.
            if cache.has_state_changed:
                self.handler_storage.file_set(cache_file, cache.serialize().encode('utf-8'))

            if permission_mode == 'delegated':
                self.connection = MSGraphAPITeamsDelegatedPermissionsClient(access_token)

            else:
                self.connection = MSGraphAPITeamsApplicationPermissionsClient(access_token)

            self._register_table('channels', ChannelsTable(self))
            self._register_table('channel_messages', ChannelMessagesTable(self))
            self._register_table('chats', ChatsTable(self))
            self._register_table('chat_messages', ChatMessagesTable(self))
            self._register_table('teams', TeamsTable(self))

        elif operation_mode == 'chatbot':
            self.connection = MicrosoftAppCredentials(
                self.connection_data['app_id'],
                self.connection_data['app_password']
            )

        else:
            raise ValueError("The supported operation modes are 'datasource' and 'chatbot'.")

        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to Microsoft Teams.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)

        try:
            connection = self.connect()
            # A connection check against the Microsoft Graph API is run if the connection is in 'datasource' mode.
            if self.connection_data.get('operation_mode', 'datasource') == 'datasource' and connection.check_connection():
                response.success = True
                response.copy_storage = True
            else:
                raise RequestException("Connection check failed!")
        except (ValueError, RequestException) as known_error:
            logger.error(f'Connection check to Microsoft Teams failed, {known_error}!')
            response.error_message = str(known_error)
        except AuthException as error:
            response.error_message = str(error)
            response.redirect_url = error.auth_url
            return response
        except Exception as unknown_error:
            logger.error(f'Connection check to Microsoft Teams failed due to an unknown error, {unknown_error}!')
            response.error_message = str(unknown_error)

        self.is_connected = response.success

        return response

    @chatbot_only
    def get_chat_config(self) -> Dict:
        """
        Gets the configuration for the chatbot.
        This method is required for the implementation of the chatbot.

        Returns:
            Dict: The configuration for the chatbot.
        """
        params = {
            'polling': {
                'type': 'webhook'
            }
        }

        return params

    @chatbot_only
    def get_my_user_name(self) -> Text:
        """
        Gets the name of the signed in user.
        This method is required for the implementation of the chatbot.

        Returns:
            Text: The name of the signed in user.
        """
        return None

    @chatbot_only
    def on_webhook(self, request: Dict, callback: Callable) -> None:
        """
        Handles a webhook request.

        Args:
            request (Dict): The request data.
            callback (Callable): The callback function to call.
        """
        self.service_url = request["serviceUrl"]
        self.channel_id = request["channelId"]
        self.bot_id = request["from"]["id"]
        self.conversation_id = request["conversation"]["id"]

        chat_bot_message = ChatBotMessage(
            ChatBotMessage.Type.DIRECT,
            text=request["text"],
            user=request["from"]["id"],
            destination=request["recipient"]["id"]
        )

        callback(
            chat_id=request['conversation']['id'],
            message=chat_bot_message
        )

    @chatbot_only
    def respond(self, message: ChatBotMessage) -> None:
        """
        Sends a response to the chatbot.

        Args:
            message (ChatBotMessage): The message to send

        Raises:
            ValueError: If the chatbot message is not of type DIRECT.

        Returns:
            None
        """
        credentials = self.connect()

        connector = ConnectorClient(credentials, base_url=self.service_url)
        connector.conversations.send_to_conversation(
            self.conversation_id,
            Activity(
                type=ActivityTypes.message,
                channel_id=self.channel_id,
                recipient=ChannelAccount(
                    id=message.destination
                ),
                from_property=ChannelAccount(
                    id=self.bot_id
                ),
                text=message.text
            )
        )



def start_http(*args, **kwargs):
    from mindsdb.utilities.log import initialize_logging

    initialize_logging("http")

    from mindsdb.api.http.start import start

    start(*args, **kwargs)


def start_mysql(*args, **kwargs):
    from mindsdb.utilities.log import initialize_logging

    initialize_logging("mysql")

    from mindsdb.api.mysql.start import start

    start(*args, **kwargs)


def start_postgres(*args, **kwargs):
    from mindsdb.utilities.log import initialize_logging

    initialize_logging("postgres")

    from mindsdb.api.postgres.start import start

    start(*args, **kwargs)


def start_tasks(*args, **kwargs):
    from mindsdb.utilities.log import initialize_logging

    initialize_logging("tasks")

    from mindsdb.interfaces.tasks.task_monitor import start

    start(*args, **kwargs)


def start_ml_task_queue(*args, **kwargs):
    from mindsdb.utilities.log import initialize_logging

    initialize_logging("ml_task_queue")

    from mindsdb.utilities.ml_task_queue.consumer import start

    start(*args, **kwargs)


def start_scheduler(*args, **kwargs):
    from mindsdb.utilities.log import initialize_logging

    initialize_logging("scheduler")

    from mindsdb.interfaces.jobs.scheduler import start

    start(*args, **kwargs)


def start_litellm(*args, **kwargs):
    """Start the LiteLLM server"""
    from mindsdb.utilities.log import initialize_logging

    initialize_logging("litellm")

    from mindsdb.api.litellm.start import start

    start(*args, **kwargs)



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __version__ as version, __description__ as description
from .connection_args import connection_args, connection_args_example
try:
    from .redshift_handler import RedshiftHandler as Handler
    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = 'Amazon Redshift'
name = 'redshift'
type = HANDLER_TYPE.DATA
icon_path = 'icon.svg'

__all__ = [
    'Handler', 'version', 'name', 'type', 'title', 'description',
    'connection_args', 'connection_args_example', 'import_error', 'icon_path'
]



from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import ERR


# base exception for known error
class SqlApiException(Exception):
    err_code = ERR.ER_SYNTAX_ERROR


class ErSqlSyntaxError(SqlApiException):
    err_code = ERR.ER_SYNTAX_ERROR


class ErWrongCharset(SqlApiException):
    err_code = ERR.ER_UNKNOWN_CHARACTER_SET



"""Utilities for knowledge base operations."""
import hashlib


def generate_document_id(content: str, content_column: str = None, provided_id: str = None) -> str:
    """
    Generate a deterministic document ID from content.
    If provided_id exists, returns it directly.
    For generated IDs, uses a short hash of just the content.

    Args:
        content: The content string
        content_column: Name of the content column (not used in ID generation, kept for backward compatibility)
        provided_id: Optional user-provided ID
    Returns:
        Deterministic document ID (either provided_id or a 16-char hash of content)
    """
    if provided_id is not None:
        return provided_id

    # Generate a shorter 16-character hash based only on content
    hash_obj = hashlib.md5(content.encode())
    return hash_obj.hexdigest()[:16]



import pandas as pd
from mindsdb.utilities import log
from mindsdb_sql_parser import ast
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions

logger = log.getLogger(__name__)


class AnnotationsTable(APITable):
    """Represents a table of annotations in Zotero."""

    def select(self, query: ast.Select):
        """Select annotations based on the provided query.

        Parameters
        ----------
        query : ast.Select
            AST (Abstract Syntax Tree) representation of the SQL query.

        Returns
        -------
        Response
            Response object containing the selected annotations as a DataFrame.
        """
        if query.where is None:  # Handle case for SELECT * FROM annotations
            df = self._get_items()
            return df[self.get_columns()]

        conditions = extract_comparison_conditions(query.where)
        supported = False  # Flag to check if the query is supported

        for op, arg1, arg2 in conditions:
            if op in {'or', 'and'}:
                raise NotImplementedError('OR and AND are not supported')
            if arg1 == 'item_id' and op == '=':
                df = self._get_item(arg2)
                supported = True
            elif arg1 == 'parent_item_id' and op == '=':
                df = self._get_item_children(arg2)
                supported = True

        if not supported:
            raise NotImplementedError('Only "item_id=" and "parent_item_id=" conditions are implemented')

        return df[self.get_columns()]

    def get_columns(self):
        """Get the columns of the annotations table.

        Returns
        -------
        list
            List of column names.
        """
        return [
            'annotationColor',
            'annotationComment',
            'annotationPageLabel',
            'annotationText',
            'annotationType',
            'dateAdded',
            'dateModified',
            'key',
            'parentItem',
            'relations',
            'tags',
            'version'
        ]

    def _get_items(self) -> pd.DataFrame:
        """Get all annotations from the Zotero API.

        Returns
        -------
        pd.DataFrame
            DataFrame containing all annotations.
        """
        if not self.handler.is_connected:
            self.handler.connect()

        try:
            method = getattr(self.handler.api, 'items')
            result = method(itemType='annotation')

            if isinstance(result, dict):
                return pd.DataFrame([result.get('data', {})])
            if isinstance(result, list) and all(isinstance(item, dict) for item in result):
                data_list = [item.get('data', {}) for item in result]
                return pd.DataFrame(data_list)

        except Exception as e:
            logger.error(f"Error fetching items: {e}")
            raise e

        return pd.DataFrame()

    def _get_item(self, item_id: str) -> pd.DataFrame:
        """Get a single annotation by item ID.

        Parameters
        ----------
        item_id : str
            The ID of the item to fetch.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the annotation.
        """
        if not self.handler.is_connected:
            self.handler.connect()

        try:
            method = getattr(self.handler.api, 'item')
            result = method(item_id, itemType='annotation')

            if isinstance(result, dict):
                return pd.DataFrame([result.get('data', {})])

        except Exception as e:
            logger.error(f"Error fetching item with ID {item_id}: {e}")
            raise e

        return pd.DataFrame()

    def _get_item_children(self, parent_item_id: str) -> pd.DataFrame:
        """Get annotations for a specific parent item ID.

        Parameters
        ----------
        parent_item_id : str
            The parent item ID to fetch annotations for.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the annotations.
        """
        if not self.handler.is_connected:
            self.handler.connect()

        try:
            method = getattr(self.handler.api, 'children')
            result = method(parent_item_id, itemType='annotation')

            if isinstance(result, dict):
                return pd.DataFrame([result.get('data', {})])
            if isinstance(result, list) and all(isinstance(item, dict) for item in result):
                data_list = [item.get('data', {}) for item in result]
                return pd.DataFrame(data_list)

        except Exception as e:
            logger.error(f"Error fetching children for parent item ID {parent_item_id}: {e}")
            raise e

        return pd.DataFrame()



"""added_metadata_to_projects

Revision ID: 11347c213b36
Revises: 4521dafe89ab
Create Date: 2025-02-19 18:46:24.014843

"""
from alembic import op
from sqlalchemy.orm.attributes import flag_modified
import sqlalchemy as sa

import mindsdb.interfaces.storage.db as db
from mindsdb.utilities.config import config


# revision identifiers, used by Alembic.
revision = '11347c213b36'
down_revision = '4521dafe89ab'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('project', schema=None) as batch_op:
        batch_op.add_column(sa.Column('metadata', sa.JSON(), nullable=True))

    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)
    session.commit()

    project = session.query(db.Project).filter_by(name='mindsdb').first()
    if project:
        project.name = config.get('default_project')
        project.metadata_ = {"is_default": True}
        flag_modified(project, 'metadata_')
        session.commit()


def downgrade():
    with op.batch_alter_table('project', schema=None) as batch_op:
        batch_op.drop_column('metadata')



from typing import List, Union
from pydantic import BaseModel, Extra, field_validator

from mindsdb.integrations.handlers.rag_handler.settings import (
    RAGBaseParameters,
)
from langchain_core.callbacks import StreamingStdOutCallbackHandler

EVAL_COLUMN_NAMES = (
    "question",
    "answers",
    "context",
)

SUPPORTED_EVALUATION_TYPES = ("retrieval", "e2e")

GENERATION_METRICS = ("rouge", "meteor", "cosine_similarity", "accuracy")
RETRIEVAL_METRICS = ("cosine_similarity", "accuracy")


# todo make a separate class for evaluation parameters


class WriterLLMParameters(BaseModel):
    """Model parameters for the Writer LLM API interface"""

    writer_api_key: str
    writer_org_id: str = None
    base_url: str = None
    model_id: str = "palmyra-x"
    callbacks: List[StreamingStdOutCallbackHandler] = [StreamingStdOutCallbackHandler()]
    max_tokens: int = 1024
    temperature: float = 0.0
    top_p: float = 1
    stop: List[str] = []
    best_of: int = 5
    verbose: bool = False

    class Config:
        extra = Extra.forbid
        arbitrary_types_allowed = True


class WriterHandlerParameters(RAGBaseParameters):
    """Model parameters for create model"""

    llm_params: WriterLLMParameters
    generation_evaluation_metrics: List[str] = list(GENERATION_METRICS)
    retrieval_evaluation_metrics: List[str] = list(RETRIEVAL_METRICS)
    evaluation_type: str = "e2e"
    n_rows_evaluation: int = None  # if None, evaluate on all rows
    retriever_match_threshold: float = 0.7
    generator_match_threshold: float = 0.8
    evaluate_dataset: Union[List[dict], str] = "squad_v2_val_100_sample"

    class Config:
        extra = Extra.forbid
        arbitrary_types_allowed = True
        use_enum_values = True

    @field_validator("generation_evaluation_metrics")
    def generation_evaluation_metrics_must_be_supported(cls, v):
        for metric in v:
            if metric not in GENERATION_METRICS:
                raise ValueError(
                    f"generation_evaluation_metrics must be one of {', '.join(str(v) for v in GENERATION_METRICS)}, got {metric}"
                )
        return v

    @field_validator("retrieval_evaluation_metrics")
    def retrieval_evaluation_metrics_must_be_supported(cls, v):
        for metric in v:
            if metric not in GENERATION_METRICS:
                raise ValueError(
                    f"retrieval_evaluation_metrics must be one of {', '.join(str(v) for v in RETRIEVAL_METRICS)}, got {metric}"
                )
        return v

    @field_validator("evaluation_type")
    def evaluation_type_must_be_supported(cls, v):
        if v not in SUPPORTED_EVALUATION_TYPES:
            raise ValueError(
                f"evaluation_type must be one of `retrieval` or `e2e`, got {v}"
            )
        return v



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    base_id={
        'type': ARG_TYPE.STR,
        'description': 'The Airtable base ID.'
    },
    table_name={
        'type': ARG_TYPE.STR,
        'description': 'The Airtable table name.'
    },
    api_key={
        'type': ARG_TYPE.STR,
        'description': 'The API key for the Airtable API.',
        'secret': True
    }
)

connection_args_example = OrderedDict(
    base_id='dqweqweqrwwqq',
    table_name='iris',
    api_key='knlsndlknslk'
)



import re
import csv
import inspect
import traceback
from io import StringIO
from typing import Iterable, List, Optional, Any, Tuple
from collections import defaultdict
import fnmatch

import pandas as pd
from mindsdb_sql_parser import parse_sql
from mindsdb_sql_parser.ast import Select, Show, Describe, Explain, Identifier

from mindsdb.utilities import log
from mindsdb.utilities.context import context as ctx
from mindsdb.integrations.utilities.query_traversal import query_traversal
from mindsdb.integrations.libs.response import INF_SCHEMA_COLUMNS_NAMES
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import MYSQL_DATA_TYPE
from mindsdb.utilities.config import config
from mindsdb.interfaces.data_catalog.data_catalog_reader import DataCatalogReader

logger = log.getLogger(__name__)


def list_to_csv_str(array: List[List[Any]]) -> str:
    """Convert a 2D array into a CSV string.

    Args:
        array (List[List[Any]]): A 2D array/list of values to convert to CSV format

    Returns:
        str: The array formatted as a CSV string using Excel dialect
    """
    output = StringIO()
    writer = csv.writer(output, dialect="excel")
    str_array = [[str(item) for item in row] for row in array]
    writer.writerows(str_array)
    return output.getvalue()


def split_table_name(table_name: str) -> List[str]:
    """Split table name from llm to parts

    Args:
        table_name (str): input table name

    Returns:
        List[str]: parts of table identifier like ['database', 'schema', 'table']

    Example:
        'input': '`aaa`.`bbb.ccc`', 'output': ['aaa', 'bbb.ccc']
        'input': '`aaa`.`bbb`.`ccc`', 'output': ['aaa', 'bbb', 'ccc']
        'input': 'aaa.bbb', 'output': ['aaa', 'bbb']
        'input': '`aaa.bbb`', 'output': ['aaa.bbb']
        'input': '`aaa.bbb.ccc`', 'output': ['aaa.bbb.ccc']
        'input': 'aaa.`bbb`', 'output': ['aaa', 'bbb']
        'input': 'aaa.bbb.ccc', 'output': ['aaa', 'bbb', 'ccc']
        'input': 'aaa.`bbb.ccc`', 'output': ['aaa', 'bbb.ccc']
        'input': '`aaa`.`bbb.ccc`', 'output': ['aaa', 'bbb.ccc']
    """
    result = []
    current = ""
    in_backticks = False

    i = 0
    while i < len(table_name):
        if table_name[i] == "`":
            in_backticks = not in_backticks
        elif table_name[i] == "." and not in_backticks:
            if current:
                result.append(current.strip("`"))
                current = ""
        else:
            current += table_name[i]
        i += 1

    if current:
        result.append(current.strip("`"))

    return result


class TablesCollection:
    """
    Collection of identifiers.
    Supports wildcard in tables name.
    """

    def __init__(self, items: List[Identifier | str] = None, default_db=None):
        if items is None:
            items = []

        self.items = items
        self._dbs = defaultdict(set)
        self._schemas = defaultdict(dict)
        self._no_db_tables = set()
        self.has_wildcard = False
        self.databases = set()
        self._default_db = default_db

        for name in items:
            if not isinstance(name, Identifier):
                name = Identifier(name)
            db, schema, tbl = self._get_paths(name)
            if db is None:
                self._no_db_tables.add(tbl)
            elif schema is None:
                self._dbs[db].add(tbl)
            else:
                if schema not in self._schemas[db]:
                    self._schemas[db][schema] = set()
                self._schemas[db][schema].add(tbl)

            if "*" in tbl:
                self.has_wildcard = True
            self.databases.add(db)

    def _get_paths(self, table: Identifier) -> Tuple:
        # split identifier to db, schema, table name
        schema = None
        db = None

        match [x.lower() for x in table.parts]:
            case [tbl]:
                pass
            case [db, tbl]:
                pass
            case [db, schema, tbl]:
                pass
            case _:
                raise NotImplementedError
        return db, schema, tbl.lower()

    def match(self, table: Identifier) -> bool:
        # Check if input table matches to tables in collection

        db, schema, tbl = self._get_paths(table)
        if db is None:
            if tbl in self._no_db_tables:
                return True
            if self._default_db is not None:
                return self.match(Identifier(parts=[self._default_db, tbl]))

        if schema is not None:
            if any([fnmatch.fnmatch(tbl, pattern) for pattern in self._schemas[db].get(schema, [])]):
                return True

        # table might be specified without schema
        return any([fnmatch.fnmatch(tbl, pattern) for pattern in self._dbs[db]])

    def __bool__(self):
        return len(self.items) > 0

    def __repr__(self):
        return f"Tables({self.items})"


class SQLAgent:
    """
    SQLAgent is a class that handles SQL queries for agents.
    """

    def __init__(
        self,
        command_executor,
        databases: List[str],
        databases_struct: dict,
        knowledge_base_database: str = "mindsdb",
        include_tables: Optional[List[str]] = None,
        ignore_tables: Optional[List[str]] = None,
        include_knowledge_bases: Optional[List[str]] = None,
        ignore_knowledge_bases: Optional[List[str]] = None,
        sample_rows_in_table_info: int = 3,
        cache: Optional[dict] = None,
    ):
        """
        Initialize SQLAgent.

        Args:
            command_executor: Executor for SQL commands
            databases (List[str]): List of databases to use
            databases_struct (dict): Dictionary of database structures
            knowledge_base_database (str): Project name where knowledge bases are stored (defaults to 'mindsdb')
            include_tables (List[str]): Tables to include
            ignore_tables (List[str]): Tables to ignore
            include_knowledge_bases (List[str]): Knowledge bases to include
            ignore_knowledge_bases (List[str]): Knowledge bases to ignore
            sample_rows_in_table_info (int): Number of sample rows to include in table info
            cache (Optional[dict]): Cache for query results
        """
        self._command_executor = command_executor
        self._mindsdb_db_struct = databases_struct
        self.knowledge_base_database = knowledge_base_database  # This is a project name, not a database connection
        self._databases = databases
        self._sample_rows_in_table_info = int(sample_rows_in_table_info)

        self._tables_to_include = TablesCollection(include_tables)
        if self._tables_to_include:
            # ignore_tables and include_tables should not be used together.
            # include_tables takes priority if it's set.
            ignore_tables = []
        self._tables_to_ignore = TablesCollection(ignore_tables)

        self._knowledge_bases_to_include = TablesCollection(include_knowledge_bases, default_db=knowledge_base_database)
        if self._knowledge_bases_to_include:
            # ignore_knowledge_bases and include_knowledge_bases should not be used together.
            # include_knowledge_bases takes priority if it's set.
            ignore_knowledge_bases = []
        self._knowledge_bases_to_ignore = TablesCollection(ignore_knowledge_bases, default_db=knowledge_base_database)

        self._cache = cache

        from mindsdb.interfaces.skills.skill_tool import SkillToolController

        # Initialize the skill tool controller from MindsDB
        self.skill_tool = SkillToolController()

    def _call_engine(self, query: str, database=None):
        # switch database
        ast_query = parse_sql(query.strip("`"))
        self._check_permissions(ast_query)

        if database is None:
            # if we use tables with prefixes it should work for any database
            if self._databases is not None:
                # if we have multiple databases, we need to check which one to use
                # for now, we will just use the first one
                database = self._databases[0] if self._databases else "mindsdb"

        ret = self._command_executor.execute_command(ast_query, database_name=database)
        return ret

    def _check_permissions(self, ast_query):
        # check type of query
        if not isinstance(ast_query, (Select, Show, Describe, Explain)):
            raise ValueError(f"Query is not allowed: {ast_query.to_string()}")

        kb_names = self.get_all_knowledge_base_names()

        # Check tables
        if self._tables_to_include:

            def _check_f(node, is_table=None, **kwargs):
                if is_table and isinstance(node, Identifier):
                    table_name = ".".join(node.parts)

                    # Check if this table is a knowledge base
                    if table_name in kb_names or node.parts[-1] in kb_names:
                        # If it's a knowledge base and we have knowledge base restrictions
                        self.check_knowledge_base_permission(node)
                    else:
                        try:
                            # Regular table check
                            self.check_table_permission(node)
                        except ValueError as origin_exc:
                            # was it badly quoted by llm?
                            if len(node.parts) == 1 and node.is_quoted[0] and "." in node.parts[0]:
                                node2 = Identifier(node.parts[0])
                                try:
                                    _check_f(node2, is_table=True)
                                    return node2
                                except ValueError:
                                    ...
                            raise origin_exc

            query_traversal(ast_query, _check_f)

    def check_knowledge_base_permission(self, node):
        if self._knowledge_bases_to_include and not self._knowledge_bases_to_include.match(node):
            raise ValueError(
                f"Knowledge base {str(node)} not found. Available knowledge bases: {', '.join(self._knowledge_bases_to_include.items)}"
            )
        # Check if it's a restricted knowledge base
        if self._knowledge_bases_to_ignore and self._knowledge_bases_to_ignore.match(node):
            raise ValueError(f"Knowledge base {str(node)} is not allowed.")

    def check_table_permission(self, node):
        if self._tables_to_include and not self._tables_to_include.match(node):
            raise ValueError(
                f"Table {str(node)} not found. Available tables: {', '.join(self._tables_to_include.items)}"
            )
        # Check if it's a restricted table
        if self._tables_to_ignore and self._tables_to_ignore.match(node):
            raise ValueError(f"Table {str(node)} is not allowed.")

    def get_usable_table_names(self) -> Iterable[str]:
        """Get a list of tables that the agent has access to.

        Returns:
            Iterable[str]: list with table names
        """
        cache_key = f"{ctx.company_id}_{','.join(self._databases)}_tables"

        # first check cache and return if found
        if self._cache:
            cached_tables = self._cache.get(cache_key)
            if cached_tables:
                return cached_tables

        if not self._tables_to_include:
            # no tables allowed
            return []
        if not self._tables_to_include.has_wildcard:
            return self._tables_to_include.items

        result_tables = []

        for db_name in self._tables_to_include.databases:
            handler = self._command_executor.session.integration_controller.get_data_handler(db_name)

            if "all" in inspect.signature(handler.get_tables).parameters:
                response = handler.get_tables(all=True)
            else:
                response = handler.get_tables()
            df = response.data_frame
            col_name = "table_name"
            if col_name not in df.columns:
                # get first column if not found
                col_name = df.columns[0]

            for _, row in df.iterrows():
                if "table_schema" in row:
                    parts = [db_name, row["table_schema"], row[col_name]]
                else:
                    parts = [db_name, row[col_name]]
                if self._tables_to_include.match(Identifier(parts=parts)):
                    if not self._tables_to_ignore.match(Identifier(parts=parts)):
                        result_tables.append(parts)

        result_tables = [".".join(x) for x in result_tables]
        if self._cache:
            self._cache.set(cache_key, set(result_tables))
        return result_tables

    def get_usable_knowledge_base_names(self) -> Iterable[str]:
        """Get a list of knowledge bases that the agent has access to.

        Returns:
            Iterable[str]: list with knowledge base names
        """

        if not self._knowledge_bases_to_include and not self._knowledge_bases_to_ignore:
            # white or black list have to be set
            return []

        # Filter knowledge bases based on ignore list
        kb_names = []
        for kb_name in self.get_all_knowledge_base_names():
            kb = Identifier(parts=[self.knowledge_base_database, kb_name])
            if self._knowledge_bases_to_include and not self._knowledge_bases_to_include.match(kb):
                continue
            if not self._knowledge_bases_to_ignore.match(kb):
                kb_names.append(kb_name)
        return kb_names

    def get_all_knowledge_base_names(self) -> Iterable[str]:
        """Get a list of all knowledge bases

        Returns:
            Iterable[str]: list with knowledge base names
        """
        # cache_key = f"{ctx.company_id}_{self.knowledge_base_database}_knowledge_bases"

        # todo we need to fix the cache, file cache can potentially store out of data information
        # # first check cache and return if found
        # if self._cache:
        #    cached_kbs = self._cache.get(cache_key)
        #     if cached_kbs:
        #        return cached_kbs

        try:
            # Query to get all knowledge bases
            ast_query = Show(category="Knowledge Bases")
            result = self._command_executor.execute_command(ast_query, database_name=self.knowledge_base_database)

            # Filter knowledge bases based on ignore list
            kb_names = []
            for row in result.data.records:
                kb_names.append(row["NAME"])

            # if self._cache:
            #     self._cache.set(cache_key, set(kb_names))

            return kb_names
        except Exception as e:
            # If there's an error, log it and return an empty list
            logger.error(f"Error in get_usable_knowledge_base_names: {str(e)}")
            return []

    def _resolve_table_names(self, table_names: List[str], all_tables: List[Identifier]) -> List[Identifier]:
        """
        Tries to find table (which comes directly from an LLM) by its name
        Handles backticks (`) and tables without databases
        """

        # index to lookup table
        tables_idx = {}
        for table in all_tables:
            # by name
            if len(table.parts) == 3:
                tables_idx[tuple(table.parts[1:])] = table
            else:
                tables_idx[(table.parts[-1],)] = table
            # by path
            tables_idx[tuple(table.parts)] = table

        tables = []
        not_found = []
        for table_name in table_names:
            if not table_name.strip():
                continue

            # Some LLMs (e.g. gpt-4o) may include backticks or quotes when invoking tools.
            table_parts = split_table_name(table_name)
            if len(table_parts) == 1:
                # most likely LLM enclosed all table name in backticks `database.table`
                table_parts = split_table_name(table_name)

            # resolved table
            table_identifier = tables_idx.get(tuple(table_parts))

            if table_identifier is None:
                not_found.append(table_name)
            else:
                tables.append(table_identifier)

        if not_found:
            raise ValueError(f"Tables: {', '.join(not_found)} not found in the database")
        return tables

    def get_knowledge_base_info(self, kb_names: Optional[List[str]] = None) -> str:
        """Get information about specified knowledge bases.
        Follows best practices as specified in: Rajkumar et al, 2022 (https://arxiv.org/abs/2204.00498)
        If `sample_rows_in_table_info`, the specified number of sample rows will be
        appended to each table description. This can increase performance as demonstrated in the paper.
        """

        kbs_info = []
        for kb in kb_names:
            key = f"{ctx.company_id}_{kb}_info"
            kb_info = self._cache.get(key) if self._cache else None
            if True or kb_info is None:
                kb_info = self.get_kb_sample_rows(kb)
                if self._cache:
                    self._cache.set(key, kb_info)

            kbs_info.append(kb_info)

        return "\n\n".join(kbs_info)

    def get_table_info(self, table_names: Optional[List[str]] = None) -> str:
        """Get information about specified tables.
        Follows best practices as specified in: Rajkumar et al, 2022 (https://arxiv.org/abs/2204.00498)
        If `sample_rows_in_table_info`, the specified number of sample rows will be
        appended to each table description. This can increase performance as demonstrated in the paper.
        """
        if config.get("data_catalog", {}).get("enabled", False):
            database_table_map = {}
            for name in table_names or self.get_usable_table_names():
                name = name.replace("`", "")

                parts = name.split(".", 1)
                # TODO: Will there be situations where parts has more than 2 elements? Like a schema?
                # This is unlikely given that we default to a single schema per database.
                if len(parts) == 1:
                    raise ValueError(f"Invalid table name: {name}. Expected format is 'database.table'.")

                database_table_map.setdefault(parts[0], []).append(parts[1])

            data_catalog_str = ""
            for database_name, table_names in database_table_map.items():
                data_catalog_reader = DataCatalogReader(database_name=database_name, table_names=table_names)

                result = data_catalog_reader.read_metadata_as_string()
                data_catalog_str += str(result or "")

            return data_catalog_str

        else:
            # TODO: Improve old logic without data catalog
            all_tables = []
            for name in self.get_usable_table_names():
                # remove backticks
                name = name.replace("`", "")

                split = name.split(".")
                if len(split) > 1:
                    all_tables.append(Identifier(parts=[split[0], split[-1]]))
                else:
                    all_tables.append(Identifier(name))

            if table_names is not None:
                all_tables = self._resolve_table_names(table_names, all_tables)

            tables_info = []
            for table in all_tables:
                key = f"{ctx.company_id}_{table}_info"
                table_info = self._cache.get(key) if self._cache else None
                if True or table_info is None:
                    table_info = self._get_single_table_info(table)
                    if self._cache:
                        self._cache.set(key, table_info)

                tables_info.append(table_info)

            return "\n\n".join(tables_info)

    def get_kb_sample_rows(self, kb_name: str) -> str:
        """Get sample rows from a knowledge base.

        Args:
            kb_name (str): The name of the knowledge base.

        Returns:
            str: A string containing the sample rows from the knowledge base.
        """
        logger.info(f"_get_sample_rows: knowledge base={kb_name}")
        command = f"select * from {kb_name} limit 10;"
        try:
            ret = self._call_engine(command)
            sample_rows = ret.data.to_lists()

            def truncate_value(val):
                str_val = str(val)
                return str_val if len(str_val) < 100 else (str_val[:100] + "...")

            sample_rows = list(map(lambda row: [truncate_value(value) for value in row], sample_rows))
            sample_rows_str = "\n" + f"{kb_name}:" + list_to_csv_str(sample_rows)
        except Exception as e:
            logger.info(f"_get_sample_rows error: {e}")
            sample_rows_str = "\n" + "\t [error] Couldn't retrieve sample rows!"

        return sample_rows_str

    def _get_single_table_info(self, table: Identifier) -> str:
        if len(table.parts) < 2:
            raise ValueError(f"Database is required for table: {table}")
        if len(table.parts) == 3:
            integration, schema_name, table_name = table.parts[-3:]
        else:
            schema_name = None
            integration, table_name = table.parts[-2:]

        table_str = str(table)

        dn = self._command_executor.session.datahub.get(integration)

        fields, dtypes = [], []
        try:
            df = dn.get_table_columns_df(table_name, schema_name)
            if not isinstance(df, pd.DataFrame) or df.empty:
                logger.warning(f"Received empty or invalid DataFrame for table columns of {table_str}")
                return f"Table named `{table_str}`:\n [No column information available]"

            fields = df[INF_SCHEMA_COLUMNS_NAMES.COLUMN_NAME].to_list()
            dtypes = [
                mysql_data_type.value if isinstance(mysql_data_type, MYSQL_DATA_TYPE) else (data_type or "UNKNOWN")
                for mysql_data_type, data_type in zip(
                    df[INF_SCHEMA_COLUMNS_NAMES.MYSQL_DATA_TYPE], df[INF_SCHEMA_COLUMNS_NAMES.DATA_TYPE]
                )
            ]
        except Exception as e:
            logger.error(f"Failed processing column info for {table_str}: {e}", exc_info=True)
            raise ValueError(f"Failed to process column info for {table_str}") from e

        if not fields:
            logger.error(f"Could not extract column fields for {table_str}.")
            return f"Table named `{table_str}`:\n [Could not extract column information]"

        try:
            sample_rows_info = self._get_sample_rows(table_str, fields)
        except Exception as e:
            logger.warning(f"Could not get sample rows for {table_str}: {e}")
            sample_rows_info = "\n\t [error] Couldn't retrieve sample rows!"

        info = f"Table named `{table_str}`:\n"
        info += f"\nSample with first {self._sample_rows_in_table_info} rows from table {table_str} in CSV format (dialect is 'excel'):\n"
        info += sample_rows_info + "\n"
        info += (
            "\nColumn data types: "
            + ",\t".join([f"\n`{field}` : `{dtype}`" for field, dtype in zip(fields, dtypes)])
            + "\n"
        )
        return info

    def _get_sample_rows(self, table: str, fields: List[str]) -> str:
        logger.info(f"_get_sample_rows: table={table} fields={fields}")
        command = f"select {', '.join(fields)} from {table} limit {self._sample_rows_in_table_info};"
        try:
            ret = self._call_engine(command)
            sample_rows = ret.data.to_lists()

            def truncate_value(val):
                str_val = str(val)
                return str_val if len(str_val) < 100 else (str_val[:100] + "...")

            sample_rows = list(map(lambda row: [truncate_value(value) for value in row], sample_rows))
            sample_rows_str = "\n" + list_to_csv_str([fields] + sample_rows)
        except Exception as e:
            logger.info(f"_get_sample_rows error: {e}")
            sample_rows_str = "\n" + "\t [error] Couldn't retrieve sample rows!"

        return sample_rows_str

    def _clean_query(self, query: str) -> str:
        # Sometimes LLM can input markdown into query tools.
        cmd = re.sub(r"```(sql)?", "", query)
        return cmd

    def query(self, command: str, fetch: str = "all") -> str:
        """Execute a SQL command and return a string representing the results.
        If the statement returns rows, a string of the results is returned.
        If the statement returns no rows, an empty string is returned.
        """

        def _repr_result(ret):
            limit_rows = 30

            columns_str = ", ".join([repr(col.name) for col in ret.columns])
            res = f"Output columns: {columns_str}\n"

            data = ret.to_lists()
            if len(data) > limit_rows:
                df = pd.DataFrame(data, columns=[col.name for col in ret.columns])

                res += f"Result has {len(data)} rows. Description of data:\n"
                res += str(df.describe(include="all")) + "\n\n"
                res += f"First {limit_rows} rows:\n"

            else:
                res += "Result in CSV format (dialect is 'excel'):\n"
            res += list_to_csv_str([[col.name for col in ret.columns]] + data[:limit_rows])
            return res

        ret = self._call_engine(self._clean_query(command))
        if fetch == "all":
            result = _repr_result(ret.data)
        elif fetch == "one":
            result = "Result in CSV format (dialect is 'excel'):\n"
            result += list_to_csv_str([[col.name for col in ret.data.columns]] + [ret.data.to_lists()[0]])
        else:
            raise ValueError("Fetch parameter must be either 'one' or 'all'")
        return str(result)

    def get_table_info_safe(self, table_names: Optional[List[str]] = None) -> str:
        try:
            logger.info(f"get_table_info_safe: {table_names}")
            return self.get_table_info(table_names)
        except Exception as e:
            logger.info(f"get_table_info_safe error: {e}")
            return f"Error: {e}"

    def query_safe(self, command: str, fetch: str = "all") -> str:
        try:
            logger.info(f"query_safe (fetch={fetch}): {command}")
            return self.query(command, fetch)
        except Exception as e:
            logger.error(f"Error in query_safe: {str(e)}\n{traceback.format_exc()}")
            logger.info(f"query_safe error: {e}")
            msg = f"Error: {e}"
            if "does not exist" in msg and " relation " in msg:
                msg += "\nAvailable tables: " + ", ".join(self.get_usable_table_names())
            return msg



from typing import AsyncIterable, Dict
from mindsdb.api.a2a.common.types import (
    SendTaskRequest,
    TaskSendParams,
    Message,
    TaskStatus,
    Artifact,
    TaskStatusUpdateEvent,
    TaskArtifactUpdateEvent,
    TaskState,
    Task,
    SendTaskResponse,
    InternalError,
    JSONRPCResponse,
    SendTaskStreamingRequest,
    SendTaskStreamingResponse,
    InvalidRequestError,
)
from mindsdb.api.a2a.common.server.task_manager import InMemoryTaskManager
from mindsdb.api.a2a.agent import MindsDBAgent
from mindsdb.api.a2a.utils import to_serializable, convert_a2a_message_to_qa_format

from typing import Union
import logging
import asyncio
import time
import traceback

logger = logging.getLogger(__name__)


def to_question_format(messages):
    """Convert A2A messages to a list of {"question": ...} dicts for agent compatibility."""
    out = []
    for msg in messages:
        if "question" in msg:
            out.append(msg)
        elif "parts" in msg and isinstance(msg["parts"], list):
            for part in msg["parts"]:
                part_dict = to_serializable(part)
                if part_dict.get("type") == "text" and "text" in part_dict:
                    out.append({"question": part_dict["text"]})
    return out


class AgentTaskManager(InMemoryTaskManager):
    def __init__(
        self,
        project_name: str,
        agent_name: str = None,
    ):
        super().__init__()
        self.project_name = project_name
        self.agent_name = agent_name
        self.tasks = {}  # Task storage
        self.lock = asyncio.Lock()  # Lock for task operations

    def _create_agent(self, user_info: Dict, agent_name: str = None) -> MindsDBAgent:
        """Create a new MindsDBAgent instance for the given agent name."""
        if not agent_name:
            raise ValueError("Agent name is required but was not provided in the request")

        return MindsDBAgent(
            agent_name=agent_name,
            project_name=self.project_name,
            user_info=user_info,
        )

    async def _stream_generator(
        self, request: SendTaskStreamingRequest, user_info: Dict
    ) -> AsyncIterable[SendTaskStreamingResponse]:
        task_send_params: TaskSendParams = request.params
        query = self._get_user_query(task_send_params)
        params = self._get_task_params(task_send_params)
        agent_name = params["agent_name"]
        streaming = params["streaming"]

        # Create and store the task first to ensure it exists
        try:
            task = await self.upsert_task(task_send_params)
            logger.info(f"Task created/updated with history length: {len(task.history) if task.history else 0}")
        except Exception as e:
            logger.error(f"Error creating task: {str(e)}")
            error_result = to_serializable(
                {
                    "id": request.id,
                    "error": to_serializable(InternalError(message=f"Error creating task: {str(e)}")),
                }
            )
            yield error_result
            return  # Early return from generator

        agent = self._create_agent(user_info, agent_name)

        # Get the history from the task object (where it was properly extracted and stored)
        history = task.history if task and task.history else []

        if not streaming:
            # If streaming is disabled, use invoke and return a single response
            try:
                result = agent.invoke(query, task_send_params.sessionId, history=history)

                # Use the parts from the agent response if available, or create them
                if "parts" in result:
                    parts = result["parts"]
                else:
                    result_text = result.get("content", "No response from MindsDB")
                    parts = [{"type": "text", "text": result_text}]

                    # Check if we have structured data
                    if "data" in result and result["data"]:
                        parts.append(
                            {
                                "type": "data",
                                "data": result["data"],
                                "metadata": {"subtype": "json"},
                            }
                        )

                # Create and yield the final response
                task_state = TaskState.COMPLETED
                artifact = Artifact(parts=parts, index=0, append=False)
                task_status = TaskStatus(state=task_state)

                # Update the task store
                await self._update_store(task_send_params.id, task_status, [artifact])

                # Yield the artifact update
                yield to_serializable(
                    SendTaskStreamingResponse(
                        id=request.id,
                        result=to_serializable(TaskArtifactUpdateEvent(id=task_send_params.id, artifact=artifact)),
                    )
                )

                # Yield the final status update
                yield to_serializable(
                    SendTaskStreamingResponse(
                        id=request.id,
                        result=to_serializable(
                            TaskStatusUpdateEvent(
                                id=task_send_params.id,
                                status=to_serializable(TaskStatus(state=task_status.state)),
                                final=True,
                            )
                        ),
                    )
                )
                return

            except Exception as e:
                logger.error(f"Error invoking agent: {e}")
                error_result = to_serializable(
                    {
                        "id": request.id,
                        "error": to_serializable(
                            JSONRPCResponse(
                                id=request.id,
                                error=to_serializable(InternalError(message=f"Error invoking agent: {str(e)}")),
                            )
                        ),
                    }
                )
                yield error_result
                return

        # If streaming is enabled (default), use the streaming implementation
        try:
            logger.debug(f"Entering agent.stream() at {time.time()}")
            # Create A2A message structure and convert using centralized utility
            a2a_message = task_send_params.message.model_dump()
            logger.debug(f"History: {history}")
            if history:
                a2a_message["history"] = [msg.model_dump() if hasattr(msg, "model_dump") else msg for msg in history]

            # Convert to Q&A format using centralized utility function
            all_messages = convert_a2a_message_to_qa_format(a2a_message)

            async for item in agent.streaming_invoke(all_messages, timeout=60):
                # Clean up: Remove verbose debug logs, keep only errors and essential info
                if isinstance(item, dict) and "artifact" in item and "parts" in item["artifact"]:
                    item["artifact"]["parts"] = [to_serializable(p) for p in item["artifact"]["parts"]]
                yield to_serializable(item)
        except Exception as e:
            logger.error(f"An error occurred while streaming the response: {e}")
            logger.error(traceback.format_exc())
            error_text = f"An error occurred while streaming the response: {str(e)}"
            # Ensure all parts are plain dicts
            parts = [{"type": "text", "text": error_text}]
            parts = [to_serializable(part) for part in parts]
            artifact = {
                "parts": parts,
                "index": 0,
                "append": False,
            }
            error_result = {
                "id": request.id,
                "error": {
                    "id": task_send_params.id,
                    "artifact": artifact,
                },
            }
            yield error_result

    async def upsert_task(self, task_send_params: TaskSendParams) -> Task:
        """Create or update a task in the task store.

        Args:
            task_send_params: The parameters for the task.

        Returns:
            The created or updated task.
        """
        logger.info(f"Upserting task {task_send_params.id}")
        async with self.lock:
            task = self.tasks.get(task_send_params.id)
            if task is None:
                # Convert the message to a dict if it's not already one
                message = task_send_params.message
                message_dict = message.dict() if hasattr(message, "dict") else message

                # Get history from request if available - check both locations
                history = []

                # First check if history is at top level (task_send_params.history)
                if hasattr(task_send_params, "history") and task_send_params.history:
                    # Convert each history item to dict if needed
                    for item in task_send_params.history:
                        item_dict = item.model_dump() if hasattr(item, "model_dump") else item
                        history.append(item_dict)
                # Also check if history is nested under message (message.history)
                elif hasattr(task_send_params.message, "history") and task_send_params.message.history:
                    for item in task_send_params.message.history:
                        item_dict = item.model_dump() if hasattr(item, "model_dump") else item
                        history.append(item_dict)

                # DO NOT add current message to history - it should be processed separately
                # The current message will be extracted during streaming from task_send_params.message

                # Create a new task
                task = Task(
                    id=task_send_params.id,
                    sessionId=task_send_params.sessionId,
                    status=TaskStatus(state=TaskState.SUBMITTED),
                    history=history,
                    artifacts=[],
                )
                self.tasks[task_send_params.id] = task
            else:
                # Convert the message to a dict if it's not already one
                message = task_send_params.message
                message_dict = message.dict() if hasattr(message, "dict") else message

                # Update the existing task
                if task.history is None:
                    task.history = []

                # If we have new history from the request, use it
                if hasattr(task_send_params, "history") and task_send_params.history:
                    # Convert each history item to dict if needed and ensure proper role
                    history = []
                    for item in task_send_params.history:
                        item_dict = item.dict() if hasattr(item, "dict") else item
                        # Ensure the role is properly set
                        if "role" not in item_dict:
                            item_dict["role"] = "assistant" if "answer" in item_dict else "user"
                        history.append(item_dict)
                    task.history = history

                # Add current message to history
                task.history.append(message_dict)
            return task

    def _validate_request(
        self, request: Union[SendTaskRequest, SendTaskStreamingRequest]
    ) -> Union[None, JSONRPCResponse]:
        """Validate the request and return an error response if invalid."""
        # Check if the request has the required parameters
        if not hasattr(request, "params") or not request.params:
            return JSONRPCResponse(
                id=request.id,
                error=InvalidRequestError(message="Missing params"),
            )

        # Check if the request has a message
        if not hasattr(request.params, "message") or not request.params.message:
            return JSONRPCResponse(
                id=request.id,
                error=InvalidRequestError(message="Missing message in params"),
            )

        # Check if the message has metadata
        if not hasattr(request.params.message, "metadata") or not request.params.message.metadata:
            return JSONRPCResponse(
                id=request.id,
                error=InvalidRequestError(message="Missing metadata in message"),
            )

        # Check if the agent name is provided in the metadata
        metadata = request.params.message.metadata
        agent_name = metadata.get("agent_name", metadata.get("agentName"))
        if not agent_name:
            return JSONRPCResponse(
                id=request.id,
                error=InvalidRequestError(
                    message="Agent name is required but was not provided in the request metadata"
                ),
            )

        return None

    async def on_send_task(self, request: SendTaskRequest, user_info: Dict) -> SendTaskResponse:
        error = self._validate_request(request)
        if error:
            return error

        return await self._invoke(request, user_info=user_info)

    async def on_send_task_subscribe(
        self, request: SendTaskStreamingRequest, user_info: Dict
    ) -> AsyncIterable[SendTaskStreamingResponse]:
        error = self._validate_request(request)
        if error:
            logger.info(f"Yielding error at {time.time()} for invalid request: {error}")
            yield to_serializable(SendTaskStreamingResponse(id=request.id, error=to_serializable(error.error)))
            return

        # We can't await an async generator directly, so we need to use it as is
        try:
            logger.debug(f"Entering streaming path at {time.time()}")
            async for response in self._stream_generator(request, user_info):
                logger.debug(f"Yielding streaming response at {time.time()} with: {str(response)[:120]}")
                yield response
        except Exception as e:
            # If an error occurs, yield an error response
            logger.error(f"Error in on_send_task_subscribe: {str(e)}")
            error_result = to_serializable(
                {
                    "id": request.id,
                    "error": to_serializable(InternalError(message=f"Error processing streaming request: {str(e)}")),
                }
            )
            yield error_result

    async def _update_store(self, task_id: str, status: TaskStatus, artifacts: list[Artifact]) -> Task:
        async with self.lock:
            try:
                task = self.tasks[task_id]
            except KeyError:
                logger.error(f"Task {task_id} not found for updating the task")
                # Create a new task with the provided ID if it doesn't exist
                # This ensures we don't fail when a task is not found
                task = Task(
                    id=task_id,
                    sessionId="recovery-session",  # Use a placeholder session ID
                    messages=[],  # No messages available
                    status=status,  # Use the provided status
                    history=[],  # No history available
                )
                self.tasks[task_id] = task

            task.status = status

            # Store assistant's response in history if we have a message
            if status.message and status.message.role == "agent":
                if task.history is None:
                    task.history = []
                # Convert message to dict if needed
                message_dict = status.message.dict() if hasattr(status.message, "dict") else status.message
                # Ensure role is set to assistant
                message_dict["role"] = "assistant"
                task.history.append(message_dict)

            if artifacts is not None:
                for artifact in artifacts:
                    if artifact.append and len(task.artifacts) > 0:
                        # Append to the last artifact
                        last_artifact = task.artifacts[-1]
                        for part in artifact.parts:
                            last_artifact.parts.append(part)
                    else:
                        # Add as a new artifact
                        task.artifacts.append(artifact)
            return task

    def _get_user_query(self, task_send_params: TaskSendParams) -> str:
        """Extract the user query from the task parameters."""
        message = task_send_params.message
        if not message.parts:
            return ""

        # Find the first text part
        for part in message.parts:
            if part.type == "text":
                return part.text

        # If no text part found, return empty string
        return ""

    def _get_task_params(self, task_send_params: TaskSendParams) -> dict:
        """Extract common parameters from task metadata."""
        metadata = task_send_params.message.metadata or {}
        # Check for both agent_name and agentName in the metadata
        agent_name = metadata.get("agent_name", metadata.get("agentName"))
        return {
            "agent_name": agent_name,
            "streaming": metadata.get("streaming", True),
            "session_id": task_send_params.sessionId,
        }

    async def _invoke(self, request: SendTaskRequest, user_info: Dict) -> SendTaskResponse:
        task_send_params: TaskSendParams = request.params
        query = self._get_user_query(task_send_params)
        params = self._get_task_params(task_send_params)
        agent_name = params["agent_name"]
        streaming = params["streaming"]
        agent = self._create_agent(user_info, agent_name)

        try:
            # Get the history from the task
            task = self.tasks.get(task_send_params.id)
            history = task.history if task and task.history else []

            # Always use streaming internally, but handle the response differently based on the streaming parameter
            all_parts = []
            final_metadata = {}

            # Create a streaming generator
            stream_gen = agent.stream(query, task_send_params.sessionId, history=history)

            if streaming:
                # For streaming mode, we'll use the streaming endpoint instead
                # Just create a minimal response to acknowledge the request
                task_state = TaskState.WORKING
                task = await self._update_store(task_send_params.id, TaskStatus(state=task_state), [])
                return to_serializable(SendTaskResponse(id=request.id, result=task))
            else:
                # For non-streaming mode, collect all chunks into a single response
                async for chunk in stream_gen:
                    # Extract parts if they exist
                    if "parts" in chunk and chunk["parts"]:
                        all_parts.extend(chunk["parts"])
                    elif "content" in chunk:
                        all_parts.append({"type": "text", "text": chunk["content"]})

                    # Extract metadata if it exists
                    if "metadata" in chunk:
                        final_metadata.update(chunk["metadata"])

                # If we didn't get any parts, create a default part
                if not all_parts:
                    all_parts = [{"type": "text", "text": "No response from MindsDB"}]

                # Create the final response
                task_state = TaskState.COMPLETED
                task = await self._update_store(
                    task_send_params.id,
                    TaskStatus(
                        state=task_state,
                        message=Message(role="agent", parts=all_parts, metadata=final_metadata),
                    ),
                    [Artifact(parts=all_parts)],
                )
                return to_serializable(SendTaskResponse(id=request.id, result=task))
        except Exception as e:
            logger.error(f"Error invoking agent: {e}")
            result_text = f"Error invoking agent: {e}"
            parts = [{"type": "text", "text": result_text}]

            task_state = TaskState.FAILED
            task = await self._update_store(
                task_send_params.id,
                TaskStatus(state=task_state, message=Message(role="agent", parts=parts)),
                [Artifact(parts=parts)],
            )
            return to_serializable(SendTaskResponse(id=request.id, result=task))



from typing import Text, Dict, Optional, Any

import oracledb
import pandas as pd
from oracledb import connect, Connection, DatabaseError, Cursor
from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)
from mindsdb.utilities import log
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import MYSQL_DATA_TYPE


oracledb.defaults.fetch_lobs = False  # Return LOBs directly as strings or bytes.
logger = log.getLogger(__name__)


def _map_type(internal_type_name: str) -> MYSQL_DATA_TYPE:
    """Map Oracle types to MySQL types.
        List of types: https://docs.oracle.com/en/database/oracle/oracle-database/23/sqlrf/Data-Types.html

    Args:
        internal_type_name (str): The name of the Oracle type to map.

    Returns:
        MYSQL_DATA_TYPE: The MySQL type that corresponds to the Oracle type.
    """
    internal_type_name = internal_type_name.upper()
    types_map = {
        (
            "VARCHAR2",
            "NVARCHAR2",
            "CHARACTER VARYING",
            "CHAR VARYING",
            "NATIONAL CHARACTER",
            "NATIONAL CHAR",
            "VARCHAR",
            "NCHAR",
            "NATIONAL CHARACTER VARYING",
            "NATIONAL CHAR VARYING",
            "NCHAR VARYING",
            "LONG VARCHAR",
        ): MYSQL_DATA_TYPE.VARCHAR,
        ("INTEGER", "INT"): MYSQL_DATA_TYPE.INT,
        ("SMALLINT",): MYSQL_DATA_TYPE.SMALLINT,
        ("NUMBER", "DECIMAL"): MYSQL_DATA_TYPE.DECIMAL,
        ("FLOAT", "BINARY_FLOAT", "REAL"): MYSQL_DATA_TYPE.FLOAT,
        ("BINARY_DOUBLE",): MYSQL_DATA_TYPE.DOUBLE,
        ("LONG",): MYSQL_DATA_TYPE.BIGINT,
        ("DATE",): MYSQL_DATA_TYPE.DATE,
        ("HOUR", "MINUTE", "SECOND", "TIMEZONE_HOUR", "TIMEZONE_MINUTE"): MYSQL_DATA_TYPE.SMALLINT,
        ("TIMESTAMP", "TIMESTAMP WITH TIME ZONE", "TIMESTAMP WITH LOCAL TIME ZONE"): MYSQL_DATA_TYPE.TIMESTAMP,
        ("RAW", "LONG RAW", "BLOB", "BFILE"): MYSQL_DATA_TYPE.BINARY,
        ("ROWID", "UROWID"): MYSQL_DATA_TYPE.TEXT,
        ("CHAR", "NCHAR", "CLOB", "NCLOB", "CHARACTER"): MYSQL_DATA_TYPE.CHAR,
        ("VECTOR",): MYSQL_DATA_TYPE.VECTOR,
        ("JSON",): MYSQL_DATA_TYPE.JSON,
    }

    for db_types_list, mysql_data_type in types_map.items():
        if internal_type_name in db_types_list:
            return mysql_data_type

    logger.debug(f"Oracle handler type mapping: unknown type: {internal_type_name}, use VARCHAR as fallback.")
    return MYSQL_DATA_TYPE.VARCHAR


def _make_table_response(result: list[tuple[Any]], cursor: Cursor) -> Response:
    """Build response from result and cursor.

    Args:
        result (list[tuple[Any]]): result of the query.
        cursor (oracledb.Cursor): cursor object.

    Returns:
        Response: response object.
    """
    description: list[tuple[Any]] = cursor.description
    mysql_types: list[MYSQL_DATA_TYPE] = []
    for column in description:
        db_type = column[1]
        precision = column[4]
        scale = column[5]
        if db_type is oracledb.DB_TYPE_JSON:
            mysql_types.append(MYSQL_DATA_TYPE.JSON)
        elif db_type is oracledb.DB_TYPE_VECTOR:
            mysql_types.append(MYSQL_DATA_TYPE.VECTOR)
        elif db_type is oracledb.DB_TYPE_NUMBER:
            if scale != 0:
                mysql_types.append(MYSQL_DATA_TYPE.FLOAT)
            else:
                # python max int is 19 digits, oracle can return more
                if precision > 18:
                    mysql_types.append(MYSQL_DATA_TYPE.DECIMAL)
                else:
                    mysql_types.append(MYSQL_DATA_TYPE.INT)
        elif db_type is oracledb.DB_TYPE_BINARY_FLOAT:
            mysql_types.append(MYSQL_DATA_TYPE.FLOAT)
        elif db_type is oracledb.DB_TYPE_BINARY_DOUBLE:
            mysql_types.append(MYSQL_DATA_TYPE.FLOAT)
        elif db_type is oracledb.DB_TYPE_BINARY_INTEGER:
            mysql_types.append(MYSQL_DATA_TYPE.INT)
        elif db_type is oracledb.DB_TYPE_BOOLEAN:
            mysql_types.append(MYSQL_DATA_TYPE.BOOLEAN)
        elif db_type in (
            oracledb.DB_TYPE_CHAR,
            oracledb.DB_TYPE_NCHAR,
            oracledb.DB_TYPE_LONG,
            oracledb.DB_TYPE_NVARCHAR,
            oracledb.DB_TYPE_VARCHAR,
            oracledb.DB_TYPE_LONG_NVARCHAR,
        ):
            mysql_types.append(MYSQL_DATA_TYPE.TEXT)
        elif db_type in (oracledb.DB_TYPE_RAW, oracledb.DB_TYPE_LONG_RAW):
            mysql_types.append(MYSQL_DATA_TYPE.BINARY)
        elif db_type is oracledb.DB_TYPE_DATE:
            mysql_types.append(MYSQL_DATA_TYPE.DATE)
        elif db_type is oracledb.DB_TYPE_TIMESTAMP:
            mysql_types.append(MYSQL_DATA_TYPE.TIMESTAMP)
        else:
            # fallback
            mysql_types.append(MYSQL_DATA_TYPE.TEXT)

    # region cast int and bool to nullable types
    serieses = []
    for i, mysql_type in enumerate(mysql_types):
        expected_dtype = None
        if mysql_type in (
            MYSQL_DATA_TYPE.SMALLINT,
            MYSQL_DATA_TYPE.INT,
            MYSQL_DATA_TYPE.MEDIUMINT,
            MYSQL_DATA_TYPE.BIGINT,
            MYSQL_DATA_TYPE.TINYINT,
        ):
            expected_dtype = "Int64"
        elif mysql_type in (MYSQL_DATA_TYPE.BOOL, MYSQL_DATA_TYPE.BOOLEAN):
            expected_dtype = "boolean"
        serieses.append(pd.Series([row[i] for row in result], dtype=expected_dtype, name=description[i][0]))
    df = pd.concat(serieses, axis=1, copy=False)
    # endregion

    return Response(RESPONSE_TYPE.TABLE, data_frame=df, mysql_types=mysql_types)


class OracleHandler(DatabaseHandler):
    """
    This handler handles connection and execution of SQL queries on Oracle.
    """

    name = "oracle"

    def __init__(self, name: Text, connection_data: Optional[Dict], **kwargs) -> None:
        """
        Initializes the handler.

        Args:
            name (Text): The name of the handler instance.
            connection_data (Dict): The connection data required to connect to Amazon DynamoDB.
            kwargs: Arbitrary keyword arguments.
        """
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def connect(self) -> Connection:
        """
        Establishes a connection to the Oracle database.

        Raises:
            ValueError: If the expected connection parameters are not provided.

        Returns:
            oracledb.Connection: A connection object to the Oracle database.
        """
        if self.is_connected is True:
            return self.connection

        # Mandatory connection parameters.
        if not all(key in self.connection_data for key in ["user", "password"]):
            raise ValueError("Required parameters (user, password) must be provided.")

        if self.connection_data.get("thick_mode", False):
            oracledb.init_oracle_client()

        config = {
            "user": self.connection_data["user"],
            "password": self.connection_data["password"],
        }

        # If 'dsn' is given, use it. Otherwise, use the individual connection parameters.
        if "dsn" in self.connection_data:
            config["dsn"] = self.connection_data["dsn"]

        else:
            if "host" not in self.connection_data and not any(
                key in self.connection_data for key in ["sid", "service_name"]
            ):
                raise ValueError(
                    "Required parameter host and either sid or service_name must be provided. Alternatively, dsn can be provided."
                )

            config["host"] = self.connection_data.get("host")

            # Optional connection parameters when 'dsn' is not given.
            optional_parameters = ["port", "sid", "service_name"]
            for parameter in optional_parameters:
                if parameter in self.connection_data:
                    config[parameter] = self.connection_data[parameter]

        # Other optional connection parameters.
        if "disable_oob" in self.connection_data:
            config["disable_oob"] = self.connection_data["disable_oob"]

        if "auth_mode" in self.connection_data:
            mode_name = "AUTH_MODE_" + self.connection_data["auth_mode"].upper()
            if not hasattr(oracledb, mode_name):
                raise ValueError(f"Unknown auth mode: {mode_name}")
            config["mode"] = getattr(oracledb, mode_name)

        try:
            connection = connect(
                **config,
            )

            if "session_variables" in self.connection_data:
                with connection.cursor() as cur:
                    for key, value in self.connection_data["session_variables"].items():
                        cur.execute(f"ALTER SESSION SET {key} = {repr(value)}")

        except DatabaseError as database_error:
            logger.error(f"Error connecting to Oracle, {database_error}!")
            raise

        except Exception as unknown_error:
            logger.error(f"Unknown error when connecting to Elasticsearch: {unknown_error}")
            raise

        self.is_connected = True
        self.connection = connection
        return self.connection

    def disconnect(self):
        """
        Closes the connection to the Oracle database if it's currently open.
        """
        if self.is_connected is False:
            return
        self.connection.close()
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the Oracle database.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            con = self.connect()
            con.ping()
            response.success = True
        except (ValueError, DatabaseError) as known_error:
            logger.error(f"Connection check to Oracle failed, {known_error}!")
            response.error_message = str(known_error)
        except Exception as unknown_error:
            logger.error(f"Connection check to Oracle failed due to an unknown error, {unknown_error}!")
            response.error_message = str(unknown_error)

        if response.success and need_to_close:
            self.disconnect()

        elif not response.success and self.is_connected:
            self.is_connected = False

        return response

    def native_query(self, query: Text) -> Response:
        """
        Executes a SQL query on the Oracle database and returns the result.

        Args:
            query (Text): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cur:
            try:
                cur.execute(query)
                if cur.description is None:
                    response = Response(RESPONSE_TYPE.OK, affected_rows=cur.rowcount)
                else:
                    result = cur.fetchall()
                    response = _make_table_response(result, cur)
                connection.commit()
            except DatabaseError as database_error:
                logger.error(f"Error running query: {query} on Oracle, {database_error}!")
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(database_error),
                )
                connection.rollback()

            except Exception as unknown_error:
                logger.error(f"Unknwon error running query: {query} on Oracle, {unknown_error}!")
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(unknown_error),
                )
                connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode and retrieves the data.

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Returns:
            Response: The response from the `native_query` method, containing the result of the SQL query execution.
        """
        renderer = SqlalchemyRender("oracle")
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Retrieves a list of all non-system tables and views in the current schema of the Oracle database.

        Returns:
            Response: A response object containing the list of tables and views, formatted as per the `Response` class.
        """
        # TODO: This query does not seem to be correct.
        query = """
            SELECT table_name
            FROM user_tables
            ORDER BY 1
        """
        return self.native_query(query)

    def get_columns(self, table_name: Text) -> Response:
        """
        Retrieves column details for a specified table in the Oracle database.

        Args:
            table_name (Text): The name of the table for which to retrieve column information.

        Returns:
            Response: A response object containing the column details, formatted as per the `Response` class.
        Raises:
            ValueError: If the 'table_name' is not a valid string.
        """
        query = f"""
            SELECT
                COLUMN_NAME,
                DATA_TYPE,
                COLUMN_ID AS ORDINAL_POSITION,
                DATA_DEFAULT AS COLUMN_DEFAULT,
                CASE NULLABLE WHEN 'Y' THEN 'YES' ELSE 'NO' END AS IS_NULLABLE,
                CHAR_LENGTH AS CHARACTER_MAXIMUM_LENGTH,
                NULL AS CHARACTER_OCTET_LENGTH,
                DATA_PRECISION AS NUMERIC_PRECISION,
                DATA_SCALE AS NUMERIC_SCALE,
                NULL AS DATETIME_PRECISION,
                CHARACTER_SET_NAME,
                NULL AS COLLATION_NAME
            FROM USER_TAB_COLUMNS
            WHERE table_name = '{table_name}'
            ORDER BY TABLE_NAME, COLUMN_ID;
        """
        result = self.native_query(query)
        if result.resp_type is RESPONSE_TYPE.TABLE:
            result.to_columns_table_response(map_type_fn=_map_type)
        return result
